# -*- coding: utf-8 -*-
"""
biosppy.clustering
------------------

This module provides various unsupervised machine learning (clustering)
algorithms.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
from six.moves import map, range, zip
import six

# 3rd party
import numpy as np
import scipy.cluster.hierarchy as sch
import scipy.cluster.vq as scv
import scipy.sparse as sp
import sklearn.cluster as skc
from sklearn.model_selection import ParameterGrid

# local
from . import metrics, utils


def dbscan(data=None,
           min_samples=5,
           eps=0.5,
           metric='euclidean',
           metric_args=None):
    """Perform clustering using the DBSCAN algorithm [EKSX96]_.

    The algorithm works by grouping data points that are closely packed
    together (with many nearby neighbors), marking as outliers points that lie
    in low-density regions.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    min_samples : int, optional
        Minimum number of samples in a cluster.
    eps : float, optional
        Maximum distance between two samples in the same cluster.
    metric : str, optional
        Distance metric (see scipy.spatial.distance).
    metric_args : dict, optional
        Additional keyword arguments to pass to the distance function.

    Returns
    -------
    clusters : dict
        Dictionary with the sample indices (rows from 'data') for each found
        cluster; outliers have key -1; clusters are assigned integer keys
        starting at 0.

    References
    ----------
    .. [EKSX96] M. Ester, H. P. Kriegel, J. Sander, and X. Xu,
       “A Density-Based Algorithm for Discovering Clusters in Large Spatial
       Databases with Noise”, Proceedings of the 2nd International
       Conf. on Knowledge Discovery and Data Mining, pp. 226-231, 1996.

    """

    # check inputs
    if data is None:
        raise TypeError("Please specify input data.")

    if metric_args is None:
        metric_args = {}

    # compute distances
    D = metrics.pdist(data, metric=metric, **metric_args)
    D = metrics.squareform(D)

    # fit
    db = skc.DBSCAN(eps=eps, min_samples=min_samples, metric='precomputed')
    labels = db.fit_predict(D)

    # get cluster indices
    clusters = _extract_clusters(labels)

    return utils.ReturnTuple((clusters,), ('clusters',))


def hierarchical(data=None,
                 k=0,
                 linkage='average',
                 metric='euclidean',
                 metric_args=None):
    """Perform clustering using hierarchical agglomerative algorithms.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    k : int, optional
        Number of clusters to extract; if 0 uses the life-time criterion.
    linkage : str, optional
        Linkage criterion; one of 'average', 'centroid', 'complete', 'median',
        'single', 'ward', or 'weighted'.
    metric : str, optional
        Distance metric (see 'biosppy.metrics').
    metric_args : dict, optional
        Additional keyword arguments to pass to the distance function.

    Returns
    -------
    clusters : dict
        Dictionary with the sample indices (rows from 'data') for each found
        cluster; outliers have key -1; clusters are assigned integer keys
        starting at 0.

    Raises
    ------
    TypeError
        If 'metric' is not a string.
    ValueError
        When the 'linkage' is unknown.
    ValueError
        When 'metric' is not 'euclidean' when using 'centroid', 'median',
        or 'ward' linkage.
    ValueError
        When 'k' is larger than the number of data samples.

    """

    # check inputs
    if data is None:
        raise TypeError("Please specify input data.")

    if linkage not in ['average', 'centroid', 'complete', 'median', 'single',
                       'ward', 'weighted']:
        raise ValueError("Unknown linkage criterion '%r'." % linkage)

    if not isinstance(metric, six.string_types):
        raise TypeError("Please specify the distance metric as a string.")

    N = len(data)
    if k > N:
        raise ValueError("Number of clusters 'k' is higher than the number" \
                          " of input samples.")

    if metric_args is None:
        metric_args = {}

    if linkage in ['centroid', 'median', 'ward']:
        if metric != 'euclidean':
            raise TypeError("Linkage '{}' requires the distance metric to be" \
                            " 'euclidean'.".format(linkage))
        Z = sch.linkage(data, method=linkage)
    else:
        # compute distances
        D = metrics.pdist(data, metric=metric, **metric_args)

        # build linkage
        Z = sch.linkage(D, method=linkage)

    if k < 0:
        k = 0

    # extract clusters
    if k == 0:
        # life-time
        labels = _life_time(Z, N)
    else:
        labels = sch.fcluster(Z, k, 'maxclust')

    # get cluster indices
    clusters = _extract_clusters(labels)

    return utils.ReturnTuple((clusters,), ('clusters',))


def kmeans(data=None,
           k=None,
           init='random',
           max_iter=300,
           n_init=10,
           tol=0.0001):
    """Perform clustering using the k-means algorithm.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    k : int
        Number of clusters to extract.
    init : str, array, optional
        If string, one of 'random' or 'k-means++'; if array, it should be of
        shape (n_clusters, n_features), specifying the initial centers.
    max_iter : int, optional
        Maximum number of iterations.
    n_init : int, optional
        Number of initializations.
    tol : float, optional
        Relative tolerance to declare convergence.

    Returns
    -------
    clusters : dict
        Dictionary with the sample indices (rows from 'data') for each found
        cluster; outliers have key -1; clusters are assigned integer keys
        starting at 0.

    """

    # check inputs
    if data is None:
        raise TypeError("Please specify input data.")

    if k is None:
        raise TypeError("Please specify the number 'k' of clusters.")

    clf = skc.KMeans(n_clusters=k,
                     init=init,
                     max_iter=max_iter,
                     n_init=n_init,
                     tol=tol)
    labels = clf.fit_predict(data)

    # get cluster indices
    clusters = _extract_clusters(labels)

    return utils.ReturnTuple((clusters,), ('clusters',))


def consensus(data=None, k=0, linkage='average', fcn=None, grid=None):
    """Perform clustering based in an ensemble of partitions.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    k : int, optional
        Number of clusters to extract; if 0 uses the life-time criterion.
    linkage : str, optional
        Linkage criterion for final partition extraction; one of 'average',
        'centroid', 'complete', 'median', 'single', 'ward', or 'weighted'.
    fcn : function
        A clustering function.
    grid : dict, list, optional
        A (list of) dictionary with parameters for each run of the clustering
        method (see sklearn.model_selection.ParameterGrid).

    Returns
    -------
    clusters : dict
        Dictionary with the sample indices (rows from 'data') for each found
        cluster; outliers have key -1; clusters are assigned integer keys
        starting at 0.

    """

    # check inputs
    if data is None:
        raise TypeError("Please specify input data.")

    if fcn is None:
        raise TypeError("Please specify the clustering function.")

    if grid is None:
        grid = {}

    # create ensemble
    ensemble, = create_ensemble(data=data, fcn=fcn, grid=grid)

    # generate coassoc
    coassoc, = create_coassoc(ensemble=ensemble, N=len(data))

    # extract partition
    clusters, = coassoc_partition(coassoc=coassoc, k=k, linkage=linkage)

    return utils.ReturnTuple((clusters,), ('clusters',))


def consensus_kmeans(data=None,
                     k=0,
                     linkage='average',
                     nensemble=100,
                     kmin=None,
                     kmax=None):
    """Perform clustering based on an ensemble of k-means partitions.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    k : int, optional
        Number of clusters to extract; if 0 uses the life-time criterion.
    linkage : str, optional
        Linkage criterion for final partition extraction; one of 'average',
        'centroid', 'complete', 'median', 'single', 'ward', or 'weighted'.
    nensemble : int, optional
        Number of partitions in the ensemble.
    kmin : int, optional
        Minimum k for the k-means partitions; defaults to :math:`\\sqrt{m}/2`.
    kmax : int, optional
        Maximum k for the k-means partitions; defaults to :math:`\\sqrt{m}`.

    Returns
    -------
    clusters : dict
        Dictionary with the sample indices (rows from 'data') for each found
        cluster; outliers have key -1; clusters are assigned integer keys
        starting at 0.

    """

    # check inputs
    if data is None:
        raise TypeError("Please specify input data.")

    N = len(data)

    if kmin is None:
        kmin = int(round(np.sqrt(N) / 2.))

    if kmax is None:
        kmax = int(round(np.sqrt(N)))

    # initialization grid
    grid = {
        'k': np.random.random_integers(low=kmin, high=kmax, size=nensemble)
    }

    # run consensus
    clusters, = consensus(data=data,
                          k=k,
                          linkage=linkage,
                          fcn=kmeans,
                          grid=grid)

    return utils.ReturnTuple((clusters,), ('clusters',))


def create_ensemble(data=None, fcn=None, grid=None):
    """Create an ensemble of partitions of the data using the given
    clustering method.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    fcn : function
        A clustering function.
    grid : dict, list, optional
        A (list of) dictionary with parameters for each run of the clustering
        method (see sklearn.model_selection.ParameterGrid).

    Returns
    -------
    ensemble : list
        Obtained ensemble partitions.

    """

    # check inputs
    if data is None:
        raise TypeError("Please specify input data.")

    if fcn is None:
        raise TypeError("Please specify the clustering function.")

    if grid is None:
        grid = {}

    # grid iterator
    grid = ParameterGrid(grid)

    # run clustering
    ensemble = []
    for params in grid:
        ensemble.append(fcn(data, **params)['clusters'])

    return utils.ReturnTuple((ensemble,), ('ensemble',))


def create_coassoc(ensemble=None, N=None):
    """Create the co-association matrix from a clustering ensemble.

    Parameters
    ----------
    ensemble : list
        Clustering ensemble partitions.
    N : int
        Number of data samples.

    Returns
    -------
    coassoc : array
        Co-association matrix.

    """

    # check inputs
    if ensemble is None:
        raise TypeError("Please specify the clustering ensemble.")

    if N is None:
        raise TypeError(
            "Please specify the number of samples in the original data set.")

    nparts = len(ensemble)
    assoc = 0
    for part in ensemble:
        nsamples = np.array([len(part[key]) for key in part])
        dim = np.sum(nsamples * (nsamples - 1)) // 2

        I = np.zeros(dim)
        J = np.zeros(dim)
        X = np.ones(dim)
        ntriplets = 0

        for v in six.itervalues(part):
            nb = len(v)
            if nb > 0:
                for h in range(nb):
                    for f in range(h + 1, nb):
                        I[ntriplets] = v[h]
                        J[ntriplets] = v[f]
                        ntriplets += 1

        assoc_aux = sp.csc_matrix((X, (I, J)), shape=(N, N))
        assoc += assoc_aux

    a = assoc + assoc.T
    a.setdiag(nparts * np.ones(N))
    coassoc = a.todense()

    return utils.ReturnTuple((coassoc,), ('coassoc',))


def coassoc_partition(coassoc=None, k=0, linkage='average'):
    """Extract the consensus partition from a co-association matrix using
    hierarchical agglomerative methods.

    Parameters
    ----------
    coassoc : array
        Co-association matrix.
    k : int, optional
        Number of clusters to extract; if 0 uses the life-time criterion.
    linkage : str, optional
        Linkage criterion for final partition extraction; one of 'average',
        'complete', 'single', or 'weighted'.

    Returns
    -------
    clusters : dict
        Dictionary with the sample indices (rows from 'data') for each found
        cluster; outliers have key -1; clusters are assigned integer keys
        starting at 0.

    """

    # check inputs
    if coassoc is None:
        raise TypeError("Please specify the input co-association matrix.")

    if linkage not in ['average', 'complete', 'single', 'weighted']:
        raise ValueError("Unknown linkage criterion '%r'." % linkage)

    N = len(coassoc)
    if k > N:
        raise ValueError("Number of clusters 'k' is higher than the number of \
                          input samples.")

    if k < 0:
        k = 0

    # convert coassoc to condensed format, dissimilarity
    mx = np.max(coassoc)
    D = metrics.squareform(mx - coassoc)

    # build linkage
    Z = sch.linkage(D, method=linkage)

    # extract clusters
    if k == 0:
        # life-time
        labels = _life_time(Z, N)
    else:
        labels = sch.fcluster(Z, k, 'maxclust')

    # get cluster indices
    clusters = _extract_clusters(labels)

    return utils.ReturnTuple((clusters,), ('clusters',))


def mdist_templates(data=None,
                    clusters=None,
                    ntemplates=1,
                    metric='euclidean',
                    metric_args=None):
    """Template selection based on the MDIST method [UlRJ04]_.

    Extends the original method with the option of also providing a data
    clustering, in which case the MDIST criterion is applied for
    each cluster [LCSF14]_.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    clusters : dict, optional
        Dictionary with the sample indices (rows from `data`) for each cluster.
    ntemplates : int, optional
        Number of templates to extract.
    metric : str, optional
        Distance metric (see scipy.spatial.distance).
    metric_args : dict, optional
        Additional keyword arguments to pass to the distance function.

    Returns
    -------
    templates : array
        Selected templates from the input data.

    References
    ----------
    .. [UlRJ04]  U. Uludag, A. Ross, A. Jain, "Biometric template selection
       and update: a case study in fingerprints",
       Pattern Recognition 37, 2004
    .. [LCSF14] A. Lourenco, C. Carreiras, H. Silva, A. Fred,
       "ECG biometrics: A template selection approach", 2014 IEEE
       International Symposium on Medical Measurements and
       Applications (MeMeA), 2014

    """

    # check inputs
    if data is None:
        raise TypeError("Please specify input data.")

    if clusters is None:
        clusters = {0: np.arange(len(data), dtype='int')}

    # cluster labels
    ks = list(clusters)

    # remove the outliers' cluster, if present
    if '-1' in ks:
        ks.remove('-1')

    cardinals = [len(clusters[k]) for k in ks]

    # check number of templates
    if np.isscalar(ntemplates):
        if ntemplates < 1:
            raise ValueError("The number of templates has to be at least 1.")
        # allocate templates per cluster
        ntemplatesPerCluster = utils.highestAveragesAllocator(cardinals,
                                                              ntemplates,
                                                              divisor='dHondt',
                                                              check=True)
    else:
        # ntemplates as a list is unofficially supported because
        # we have to account for cluster label order
        if np.sum(ntemplates) < 1:
            raise ValueError(
                "The total number of templates has to be at least 1.")
        # just copy
        ntemplatesPerCluster = ntemplates

    templates = []

    for i, k in enumerate(ks):
        c = np.array(clusters[k])
        length = cardinals[i]
        nt = ntemplatesPerCluster[i]

        if nt == 0:
            continue

        if length == 0:
            continue
        elif length == 1:
            templates.append(data[c][0])
        elif length == 2:
            if nt == 1:
                # choose randomly
                r = round(np.random.rand())
                templates.append(data[c][r])
            else:
                for j in range(length):
                    templates.append(data[c][j])
        else:
            # compute mean distances
            indices, _ = _mean_distance(data[c],
                                        metric=metric,
                                        metric_args=metric_args)

            # select templates
            sel = indices[:nt]
            for item in sel:
                templates.append(data[c][item])

    templates = np.array(templates)

    return utils.ReturnTuple((templates,), ('templates',))


def centroid_templates(data=None, clusters=None, ntemplates=1):
    """Template selection based on cluster centroids.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    clusters : dict
        Dictionary with the sample indices (rows from 'data') for each cluster.
    ntemplates : int, optional
        Number of templates to extract; if more than 1, k-means is used to
        obtain more templates.

    Returns
    -------
    templates : array
        Selected templates from the input data.

    """

    # check inputs
    if data is None:
        raise TypeError("Please specify input data.")

    if clusters is None:
        raise TypeError("Please specify a data clustering.")

    # cluster labels
    ks = list(clusters)

    # remove the outliers' cluster, if present
    if '-1' in ks:
        ks.remove('-1')

    cardinals = [len(clusters[k]) for k in ks]

    # check number of templates
    if np.isscalar(ntemplates):
        if ntemplates < 1:
            raise ValueError("The number of templates has to be at least 1.")
        # allocate templates per cluster
        ntemplatesPerCluster = utils.highestAveragesAllocator(cardinals,
                                                              ntemplates,
                                                              divisor='dHondt',
                                                              check=True)
    else:
        # ntemplates as a list is unofficially supported because
        # we have to account for cluster label order
        if np.sum(ntemplates) < 1:
            raise ValueError(
                "The total number of templates has to be at least 1.")
        # just copy
        ntemplatesPerCluster = ntemplates

    # select templates
    templates = []
    for i, k in enumerate(ks):
        c = np.array(clusters[k])
        length = cardinals[i]
        nt = ntemplatesPerCluster[i]

        # ignore cases
        if nt == 0 or length == 0:
            continue

        if nt == 1:
            # cluster centroid
            templates.append(np.mean(data[c], axis=0))
        elif nt == length:
            # centroids are the samples
            templates.extend(data[c])
        else:
            # divide space using k-means
            nb = min([nt, length])
            centroidsKmeans, _ = scv.kmeans2(data[c],
                                             k=nb,
                                             iter=50,
                                             minit='points')
            for item in centroidsKmeans:
                templates.append(item)

    templates = np.array(templates)

    return utils.ReturnTuple((templates,), ('templates',))


def outliers_dbscan(data=None,
                    min_samples=5,
                    eps=0.5,
                    metric='euclidean',
                    metric_args=None):
    """Perform outlier removal using the DBSCAN algorithm.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    min_samples : int, optional
        Minimum number of samples in a cluster.
    eps : float, optional
        Maximum distance between two samples in the same cluster.
    metric : str, optional
        Distance metric (see scipy.spatial.distance).
    metric_args : dict, optional
        Additional keyword arguments to pass to the distance function.

    Returns
    -------
    clusters : dict
        Dictionary with the sample indices (rows from 'data') for the
        outliers (key -1) and the normal (key 0) groups.
    templates : dict
        Elements from 'data' for the outliers (key -1) and the
        normal (key 0) groups.

    """

    # perform clustering
    clusters, = dbscan(data=data,
                       min_samples=min_samples,
                       eps=eps,
                       metric=metric,
                       metric_args=metric_args)

    # merge clusters
    clusters = _merge_clusters(clusters)

    # separate templates
    templates = {-1: data[clusters[-1]], 0: data[clusters[0]]}

    # output
    args = (clusters, templates)
    names = ('clusters', 'templates')

    return utils.ReturnTuple(args, names)


def outliers_dmean(data=None,
                   alpha=0.5,
                   beta=1.5,
                   metric='euclidean',
                   metric_args=None,
                   max_idx=None):
    """Perform outlier removal using the DMEAN algorithm [LCSF13]_.

    A sample is considered valid if it cumulatively verifies:
        * distance to average template smaller than a (data derived)
          threshold 'T';
        * sample minimum greater than a (data derived) threshold 'M';
        * sample maximum smaller than a (data derived) threshold 'N';
        * position of the sample maximum is the same as the
          given index [optional].

    For a set of :math:`\\{X_1, ..., X_n\\}` :math:`n` samples:

    .. math::

        \\widetilde{X} = \\frac{1}{n} \\sum_{i=1}^{n}{X_i}

        d_i = dist(X_i, \\widetilde{X})

        D_m = \\frac{1}{n} \\sum_{i=1}^{n}{d_i}

        D_s = \\sqrt{\\frac{1}{n - 1} \\sum_{i=1}^{n}{(d_i - D_m)^2}}

        T = D_m + \\alpha * D_s

        M = \\beta * median(\\{\\max{X_i}, i=1, ..., n \\})

        N = \\beta * median(\\{\\min{X_i}, i=1, ..., n \\})

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    alpha : float, optional
        Parameter for the distance threshold.
    beta : float, optional
        Parameter for the maximum and minimum thresholds.
    metric : str, optional
        Distance metric (see scipy.spatial.distance).
    metric_args : dict, optional
        Additional keyword arguments to pass to the distance function.
    max_idx : int, optional
        Index of the expected maximum.

    Returns
    -------
    clusters : dict
        Dictionary with the sample indices (rows from 'data') for the
        outliers (key -1) and the normal (key 0) groups.
    templates : dict
        Elements from 'data' for the outliers (key -1) and the
        normal (key 0) groups.

    References
    ----------
    .. [LCSF13] A. Lourenco, H. Silva, C. Carreiras, A. Fred, "Outlier
       Detection in Non-intrusive ECG Biometric System", Image Analysis
       and Recognition, vol. 7950, pp. 43-52, 2013

    """

    # check inputs
    if data is None:
        raise TypeError("Please specify input data.")

    if metric_args is None:
        metric_args = {}

    # distance to mean wave
    mean_wave = np.mean(data, axis=0, keepdims=True)
    dists = metrics.cdist(data, mean_wave, metric=metric, **metric_args)
    dists = dists.flatten()

    # distance threshold
    th = np.mean(dists) + alpha * np.std(dists, ddof=1)

    # median of max and min
    M = np.median(np.max(data, 1)) * beta
    m = np.median(np.min(data, 1)) * beta

    # search for outliers
    outliers = []
    for i, item in enumerate(data):
        idx = np.argmax(item)
        if (max_idx is not None) and (idx != max_idx):
            outliers.append(i)
        elif item[idx] > M:
            outliers.append(i)
        elif np.min(item) < m:
            outliers.append(i)
        elif dists[i] > th:
            outliers.append(i)

    outliers = np.unique(outliers)
    normal = np.setdiff1d(list(range(len(data))), outliers, assume_unique=True)

    # output
    clusters = {-1: outliers, 0: normal}

    templates = {-1: data[outliers], 0: data[normal]}

    args = (clusters, templates)
    names = ('clusters', 'templates')

    return utils.ReturnTuple(args, names)


def _life_time(Z, N):
    """Life-Time criterion for automatic selection of the number of clusters.

    Parameters
    ----------
    Z : array
        The hierarchical clustering encoded as a linkage matrix.
    N : int
        Number of data samples.

    Returns
    -------
    labels : array
        Cluster labels.

    """

    if N < 3:
        return np.arange(N, dtype='int')

    # find maximum
    df = np.diff(Z[:, 2])
    idx = np.argmax(df)
    mx = df[idx]
    th = Z[idx, 2]

    idxs = Z[np.nonzero(Z[:, 2] > th)[0], 2]
    cont = len(idxs) + 1

    # find minimum
    mi = np.min(df[np.nonzero(df != 0)])

    if mi != mx:
        if mx < 2 * mi:
            cont = 1

    if cont > 1:
        labels = sch.fcluster(Z, cont, 'maxclust')
    else:
        labels = np.arange(N, dtype='int')

    return labels


def _extract_clusters(labels):
    """Extract cluster indices from an array of cluster labels.

    Parameters
    ----------
    labels : array
        Input cluster labels.

    Returns
    -------
    clusters : dict
        Dictionary with the sample indices for each found cluster; outliers
        have key -1; clusters are assigned integer keys starting at 0.

    """

    # ensure numpy
    labels = np.array(labels)

    # unique labels and sort
    unq = np.unique(labels).tolist()

    clusters = {}

    # outliers
    if -1 in unq:
        clusters[-1] = np.nonzero(labels == -1)[0]
        unq.remove(-1)
    elif '-1' in unq:
        clusters[-1] = np.nonzero(labels == '-1')[0]
        unq.remove('-1')

    for i, u in enumerate(unq):
        clusters[i] = np.nonzero(labels == u)[0]

    return clusters


def _mean_distance(data, metric='euclidean', metric_args=None):
    """Compute the sorted mean distance between the input samples.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    metric : str, optional
        Distance metric (see scipy.spatial.distance).
    metric_args : dict, optional
        Additional keyword arguments to pass to the distance function.

    Returns
    -------
    indices : array
        Indices that sort the computed mean distances.
    mdist : array
        Mean distance characterizing each data sample.

    """

    if metric_args is None:
        metric_args = {}

    # compute distances
    D = metrics.pdist(data, metric=metric, **metric_args)
    D = metrics.squareform(D)

    # compute mean
    mdist = np.mean(D, axis=0)

    # sort
    indices = np.argsort(mdist)

    return indices, mdist


def _merge_clusters(clusters):
    """Merge non-outlier clusters in a partition.

    Parameters
    ----------
    clusters : dict
        Dictionary with the sample indices for each found cluster;
        outliers have key -1.

    Returns
    -------
    res : dict
        Merged clusters.

    """

    keys = list(clusters)

    # outliers
    if -1 in keys:
        keys.remove(-1)
        res = {-1: clusters[-1]}
    else:
        res = {-1: np.array([], dtype='int')}

    # normal clusters
    aux = np.concatenate([clusters[k] for k in keys])
    res[0] = np.unique(aux).astype('int')

    return res



# -*- coding: utf-8 -*-
"""
biosppy.stats
---------------

This module provides statistica functions and related tools.

:copyright: (c) 2015-2021 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
import six

# local

from . import utils

# 3rd party
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import pearsonr, ttest_rel, ttest_ind


def pearson_correlation(x=None, y=None):
    """Compute the Pearson Correlation Coefficient between two signals.

    The coefficient is given by:

    .. math::

        r_{xy} = \\frac{E[(X - \\mu_X) (Y - \\mu_Y)]}{\\sigma_X \\sigma_Y}

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.

    Returns
    -------
    r : float
        Pearson correlation coefficient, ranging between -1 and +1.
    pvalue : float
        Two-tailed p-value. The p-value roughly indicates the probability of
        an uncorrelated system producing datasets that have a Pearson correlation
        at least as extreme as the one computed from these datasets.

    Raises
    ------
    ValueError
        If the input signals do not have the same length.

    """

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    # ensure numpy
    x = np.array(x)
    y = np.array(y)

    n = len(x)

    if n != len(y):
        raise ValueError("Input signals must have the same length.")

    r, pvalue = pearsonr(x, y)

    return r, pvalue


def linear_regression(x=None, y=None):
    """Plot the linear regression between two signals and get the equation coefficients.

    The linear regression uses the least squares method.

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.

    Returns
    -------
    coeffs : array
        Linear regression coefficients: [m, b].

    Raises
    ------
    ValueError
        If the input signals do not have the same length.

    """

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    # ensure numpy
    x = np.array(x)
    y = np.array(y)

    n = len(x)

    if n != len(y):
        raise ValueError("Input signals must have the same length.")

    coeffs = np.polyfit(x, y, 1)
    f = np.poly1d(coeffs)

    x_min = x.min()
    x_max = x.max()

    y_min = f(x_min)
    y_max = f(x_max)

    plt.scatter(x, y)
    plt.plot(
        [x_min, x_max],
        [y_min, y_max],
        c="orange",
        label="y={:.3f}x+{:.3f}".format(coeffs[0], coeffs[1]),
    )
    plt.title("Linear Regression")
    plt.xlabel("x")
    plt.ylabel("y")
    plt.legend()

    return coeffs


def paired_test(x=None, y=None):
    """
    Perform the Student's paired t-test on the arrays x and y.
    This is a two-sided test for the null hypothesis that 2 related
    or repeated samples have identical average (expected) values.

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.

    Returns
    -------
    statistic : float
        t-statistic. The t-statistic is used in a t-test to determine
        if you should support or reject the null hypothesis.
    pvalue : float
        Two-sided p-value.

    Raises
    ------
    ValueError
        If the input signals do not have the same length.

    """

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    # ensure numpy
    x = np.array(x)
    y = np.array(y)

    n = len(x)

    if n != len(y):
        raise ValueError("Input signals must have the same length.")

    statistic, pvalue = ttest_rel(x, y)

    return statistic, pvalue


def unpaired_test(x=None, y=None):
    """
    Perform the Student's unpaired t-test on the arrays x and y.
    This is a two-sided test for the null hypothesis that 2 independent
    samples have identical average (expected) values. This test assumes
    that the populations have identical variances by default.

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.

    Returns
    -------
    statistic : float
        t-statistic. The t-statistic is used in a t-test to determine
        if you should support or reject the null hypothesis.
    pvalue : float
        Two-sided p-value.

    Raises
    ------
    ValueError
        If the input signals do not have the same length.

    """

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    # ensure numpy
    x = np.array(x)
    y = np.array(y)

    n = len(x)

    if n != len(y):
        raise ValueError("Input signals must have the same length.")

    statistic, pvalue = ttest_ind(x, y)

    return statistic, pvalue



# -*- coding: utf-8 -*-
"""
biosppy.signals
---------------

This package provides methods to synthesize common
physiological signals (biosignals):
    * Electrocardiogram (ECG)

:copyright: (c) 2015-2021 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# compat
from __future__ import absolute_import, division, print_function

# allow lazy loading
from . import ecg



# -*- coding: utf-8 -*-
"""
biosppy.signals.acc
-------------------

This module provides methods to process Acceleration (ACC) signals.
Implemented code assumes ACC acquisition from a 3 orthogonal axis reference system.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.

Authors
-------
Afonso Ferreira
Diogo Vieira

"""

# Imports
from __future__ import absolute_import, division, print_function
from six.moves import range

# 3rd party
import numpy as np

# local
from .. import plotting, utils


def acc(signal=None, sampling_rate=100., path=None, show=True):
    """Process a raw ACC signal and extract relevant signal features using
    default parameters.

    Parameters
    ----------
    signal : array
        Raw ACC signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show a summary plot.

    Returns
    -------
    ts : array
        Signal time axis reference (seconds).
    signal : array
        Raw (unfiltered) ACC signal.
    vm : array
        Vector Magnitude feature of the signal
    sm : array
        Signal Magnitude feature of the signal

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    sampling_rate = float(sampling_rate)

    # extract features
    vm_features, sm_features = extract_acc_features(signal=signal,
                                                    sampling_rate=sampling_rate)

    # get time vectors
    length = len(signal)
    T = (length - 1) / sampling_rate
    ts = np.linspace(0, T, length, endpoint=True)

    # plot
    if show:
        plotting.plot_acc(ts=ts,
                          raw=signal,
                          vm=vm_features,
                          sm=sm_features,
                          path=path,
                          show=True)

    # output
    args = (ts, signal, vm_features, sm_features)
    names = ('ts', 'signal', 'vm', 'sm')

    return utils.ReturnTuple(args, names)


def extract_acc_features(signal=None, sampling_rate=100.0):
    """Extracts the vector magnitude and signal magnitude features from an input ACC signal, given the signal itself.

    Parameters
    ----------
    signal : array
        Input ACC signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).

    Returns
    -------
    vm_features : array
        Extracted Vector Magnitude (VM) feature.
    sm_features : array
        Extracted Signal Magnitude (SM) feature.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # get acceleration features
    vm_features = np.zeros(signal.shape[0])
    sm_features = np.zeros(signal.shape[0])

    for i in range(signal.shape[0]):
        vm_features[i] = np.linalg.norm(np.array([signal[i][0], signal[i][1], signal[i][2]]))
        sm_features[i] = (abs(signal[i][0]) + abs(signal[i][1]) + abs(signal[i][2])) / 3

    return utils.ReturnTuple((vm_features, sm_features), ('vm', 'sm'))




# -*- coding: utf-8 -*-
"""
biosppy.signals.emg
-------------------

This module provides methods to process Electromyographic (EMG) signals.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function

# 3rd party
import numpy as np

# local
from . import tools as st
from .. import plotting, utils


def emg(signal=None, sampling_rate=1000., path=None, show=True):
    """Process a raw EMG signal and extract relevant signal features using
    default parameters.

    Parameters
    ----------
    signal : array
        Raw EMG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show a summary plot.

    Returns
    -------
    ts : array
        Signal time axis reference (seconds).
    filtered : array
        Filtered EMG signal.
    onsets : array
        Indices of EMG pulse onsets.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    sampling_rate = float(sampling_rate)

    # filter signal
    filtered, _, _ = st.filter_signal(signal=signal,
                                      ftype='butter',
                                      band='highpass',
                                      order=4,
                                      frequency=100,
                                      sampling_rate=sampling_rate)

    # find onsets
    onsets, = find_onsets(signal=filtered, sampling_rate=sampling_rate)

    # get time vectors
    length = len(signal)
    T = (length - 1) / sampling_rate
    ts = np.linspace(0, T, length, endpoint=True)

    # plot
    if show:
        plotting.plot_emg(ts=ts,
                          sampling_rate=1000.,
                          raw=signal,
                          filtered=filtered,
                          processed=None,
                          onsets=onsets,
                          path=path,
                          show=True)

    # output
    args = (ts, filtered, onsets)
    names = ('ts', 'filtered', 'onsets')

    return utils.ReturnTuple(args, names)


def find_onsets(signal=None, sampling_rate=1000., size=0.05, threshold=None):
    """Determine onsets of EMG pulses.

    Skips corrupted signal parts.

    Parameters
    ----------
    signal : array
        Input filtered EMG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    size : float, optional
        Detection window size (seconds).
    threshold : float, optional
        Detection threshold.

    Returns
    -------
    onsets : array
        Indices of EMG pulse onsets.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # full-wave rectification
    fwlo = np.abs(signal)

    # smooth
    size = int(sampling_rate * size)
    mvgav, _ = st.smoother(signal=fwlo,
                           kernel='boxzen',
                           size=size,
                           mirror=True)

    # threshold
    if threshold is None:
        aux = np.abs(mvgav)
        threshold = 1.2 * np.mean(aux) + 2.0 * np.std(aux, ddof=1)

    # find onsets
    length = len(signal)
    start = np.nonzero(mvgav > threshold)[0]
    stop = np.nonzero(mvgav <= threshold)[0]

    onsets = np.union1d(np.intersect1d(start - 1, stop),
                        np.intersect1d(start + 1, stop))

    if np.any(onsets):
        if onsets[-1] >= length:
            onsets[-1] = length - 1

    return utils.ReturnTuple((onsets,), ('onsets',))


def hodges_bui_onset_detector(signal=None, rest=None, sampling_rate=1000.,
                              size=None, threshold=None):
    """Determine onsets of EMG pulses.

    Follows the approach by Hodges and Bui [HoBu96]_.

    Parameters
    ----------
    signal : array
        Input filtered EMG signal.
    rest : array, list, dict
        One of the following 3 options:
        * N-dimensional array with filtered samples corresponding to a
        rest period;
        * 2D array or list with the beginning and end indices of a segment of
        the signal corresponding to a rest period;
        * Dictionary with {'mean': mean value, 'std_dev': standard variation}.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    size : int
        Detection window size (seconds).
    threshold : int, float
        Detection threshold.

    Returns
    -------
    onsets : array
        Indices of EMG pulse onsets.
    processed : array
        Processed EMG signal.

    References
    ----------
    .. [HoBu96] Hodges PW, Bui BH, "A comparison of computer-based methods for
       the determination of onset of muscle contraction using
       electromyography", Electroencephalography and Clinical Neurophysiology 
       - Electromyography and Motor Control, vol. 101:6, pp. 511-519, 1996

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if rest is None:
        raise TypeError("Please specidy rest parameters.")

    if size is None:
        raise TypeError("Please specify the detection window size.")

    if threshold is None:
        raise TypeError("Please specify the detection threshold.")

    # gather statistics on rest signal
    if isinstance(rest, np.ndarray) or isinstance(rest, list):
        # if the input parameter is a numpy array or a list
        if len(rest) >= 2:
            # first ensure numpy
            rest = np.array(rest)
            if len(rest) == 2:
                # the rest signal is a segment of the signal
                rest_signal = signal[rest[0]:rest[1]]
            else:
                # the rest signal is provided as is
                rest_signal = rest
            rest_zero_mean = rest_signal - np.mean(rest_signal)
            statistics = st.signal_stats(signal=rest_zero_mean)
            mean_rest = statistics['mean']
            std_dev_rest = statistics['std_dev']
        else:
            raise TypeError("Please specify the rest analysis.")
    elif isinstance(rest, dict):
        # if the input is a dictionary
        mean_rest = rest['mean']
        std_dev_rest = rest['std_dev']
    else:
        raise TypeError("Please specify the rest analysis.")

    # subtract baseline offset
    signal_zero_mean = signal - np.mean(signal)

    # full-wave rectification
    fwlo = np.abs(signal_zero_mean)

    # moving average
    mvgav = np.convolve(fwlo, np.ones((size,))/size, mode='valid')

    # calculate the test function
    tf = (1 / std_dev_rest) * (mvgav - mean_rest)

    # find onsets
    length = len(signal)
    start = np.nonzero(tf >= threshold)[0]
    stop = np.nonzero(tf < threshold)[0]

    onsets = np.union1d(np.intersect1d(start - 1, stop),
                        np.intersect1d(start + 1, stop))

    # adjust indices because of moving average
    onsets += int(size / 2)

    if np.any(onsets):
        if onsets[-1] >= length:
            onsets[-1] = length - 1

    return utils.ReturnTuple((onsets, tf), ('onsets', 'processed'))


def bonato_onset_detector(signal=None, rest=None, sampling_rate=1000.,
                          threshold=None, active_state_duration=None,
                          samples_above_fail=None, fail_size=None):
    """Determine onsets of EMG pulses.

    Follows the approach by Bonato et al. [Bo98]_.

    Parameters
    ----------
    signal : array
        Input filtered EMG signal.
    rest : array, list, dict
        One of the following 3 options:
        * N-dimensional array with filtered samples corresponding to a
        rest period;
        * 2D array or list with the beginning and end indices of a segment of
        the signal corresponding to a rest period;
        * Dictionary with {'mean': mean value, 'std_dev': standard variation}.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    threshold : int, float
        Detection threshold.
    active_state_duration: int
        Minimum duration of the active state.
    samples_above_fail : int
        Number of samples above the threshold level in a group of successive
        samples.
    fail_size : int
        Number of successive samples.

    Returns
    -------
    onsets : array
        Indices of EMG pulse onsets.
    processed : array
        Processed EMG signal.

    References
    ----------
    .. [Bo98] Bonato P, D’Alessio T, Knaflitz M, "A statistical method for the
       measurement of muscle activation intervals from surface myoelectric
       signal during gait", IEEE Transactions on Biomedical Engineering,
       vol. 45:3, pp. 287–299, 1998

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if rest is None:
        raise TypeError("Please specidy rest parameters.")

    if threshold is None:
        raise TypeError("Please specify the detection threshold.")

    if active_state_duration is None:
        raise TypeError("Please specify the mininum duration of the "
                        "active state.")

    if samples_above_fail is None:
        raise TypeError("Please specify the number of samples above the "
                        "threshold level in a group of successive samples.")

    if fail_size is None:
        raise TypeError("Please specify the number of successive samples.")

    # gather statistics on rest signal
    if isinstance(rest, np.ndarray) or isinstance(rest, list):
        # if the input parameter is a numpy array or a list
        if len(rest) >= 2:
            # first ensure numpy
            rest = np.array(rest)
            if len(rest) == 2:
                # the rest signal is a segment of the signal
                rest_signal = signal[rest[0]:rest[1]]
            else:
                # the rest signal is provided as is
                rest_signal = rest
            rest_zero_mean = rest_signal - np.mean(rest_signal)
            statistics = st.signal_stats(signal=rest_zero_mean)
            var_rest = statistics['var']
        else:
            raise TypeError("Please specify the rest analysis.")
    elif isinstance(rest, dict):
        # if the input is a dictionary
        var_rest = rest['var']
    else:
        raise TypeError("Please specify the rest analysis.")

    # subtract baseline offset
    signal_zero_mean = signal - np.mean(signal)

    tf_list = []
    onset_time_list = []
    offset_time_list = []
    alarm_time = 0
    state_duration = 0
    j = 0
    n = 0
    onset = False
    alarm = False
    for k in range(1, len(signal_zero_mean), 2):  # odd values only
        # calculate the test function
        tf = (1 / var_rest) * (signal_zero_mean[k-1]**2 + signal_zero_mean[k]**2)
        tf_list.append(tf)
        if onset is True:
            if alarm is False:
                if tf < threshold:
                    alarm_time = k // 2
                    alarm = True
            else:  # now we have to check for the remaining rule to me bet - duration of inactive state
                if tf < threshold:
                    state_duration += 1
                    if j > 0:  # there was one (or more) samples above the threshold level but now one is bellow it
                        # the test function may go above the threshold , but each time not longer than j samples
                        n += 1
                        if n == samples_above_fail:
                            n = 0
                            j = 0
                    if state_duration == active_state_duration:
                        offset_time_list.append(alarm_time)
                        onset = False
                        alarm = False
                        n = 0
                        j = 0
                        state_duration = 0
                else:  # sample falls below the threshold level
                    j += 1
                    if j > fail_size:
                        # the inactive state is above the threshold for longer than the predefined number of samples
                        alarm = False
                        n = 0
                        j = 0
                        state_duration = 0
        else:  # we only look for another onset if a previous offset was detected
            if alarm is False:  # if the alarm time has not yet been identified
                if tf >= threshold:  # alarm time
                    alarm_time = k // 2
                    alarm = True
            else:  # now we have to check for the remaining rule to me bet - duration of active state
                if tf >= threshold:
                    state_duration += 1
                    if j > 0:  # there was one (or more) samples below the threshold level but now one is above it.
                        # a total of n samples must be above it
                        n += 1
                        if n == samples_above_fail:
                            n = 0
                            j = 0
                    if state_duration == active_state_duration:
                        onset_time_list.append(alarm_time)
                        onset = True
                        alarm = False
                        n = 0
                        j = 0
                        state_duration = 0
                else:  # sample falls below the threshold level
                    j += 1
                    if j > fail_size:
                        # the active state has fallen below the threshold for longer than the predefined number of samples
                        alarm = False
                        n = 0
                        j = 0
                        state_duration = 0

    onsets = np.union1d(onset_time_list,
                        offset_time_list)

    # adjust indices because of odd numbers
    onsets *= 2

    return utils.ReturnTuple((onsets, tf_list), ('onsets', 'processed'))


def lidierth_onset_detector(signal=None, rest=None, sampling_rate=1000.,
                            size=None, threshold=None,
                            active_state_duration=None, fail_size=None):
    """Determine onsets of EMG pulses.

    Follows the approach by Lidierth. [Li86]_.

    Parameters
    ----------
    signal : array
        Input filtered EMG signal.
    rest : array, list, dict
        One of the following 3 options:
        * N-dimensional array with filtered samples corresponding to a
        rest period;
        * 2D array or list with the beginning and end indices of a segment of
        the signal corresponding to a rest period;
        * Dictionary with {'mean': mean value, 'std_dev': standard variation}.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    size : int
        Detection window size (seconds).
    threshold : int, float
        Detection threshold.
    active_state_duration: int
        Minimum duration of the active state.
    fail_size : int
        Number of successive samples.

    Returns
    -------
    onsets : array
        Indices of EMG pulse onsets.
    processed : array
        Processed EMG signal.

    References
    ----------
    .. [Li86] Lidierth M, "A computer based method for automated measurement
       of the periods of muscular activity from an EMG and its application to
       locomotor EMGs", ElectroencephClin Neurophysiol, vol. 64:4,
       pp. 378–380, 1986

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if rest is None:
        raise TypeError("Please specidy rest parameters.")

    if size is None:
        raise TypeError("Please specify the detection window size.")

    if threshold is None:
        raise TypeError("Please specify the detection threshold.")

    if active_state_duration is None:
        raise TypeError("Please specify the mininum duration of the "
                        "active state.")

    if fail_size is None:
        raise TypeError("Please specify the number of successive samples.")

    # gather statistics on rest signal
    if isinstance(rest, np.ndarray) or isinstance(rest, list):
        # if the input parameter is a numpy array or a list
        if len(rest) >= 2:
            # first ensure numpy
            rest = np.array(rest)
            if len(rest) == 2:
                # the rest signal is a segment of the signal
                rest_signal = signal[rest[0]:rest[1]]
            else:
                # the rest signal is provided as is
                rest_signal = rest
            rest_zero_mean = rest_signal - np.mean(rest_signal)
            statistics = st.signal_stats(signal=rest_zero_mean)
            mean_rest = statistics['mean']
            std_dev_rest = statistics['std_dev']
        else:
            raise TypeError("Please specify the rest analysis.")
    elif isinstance(rest, dict):
        # if the input is a dictionary
        mean_rest = rest['mean']
        std_dev_rest = rest['std_dev']
    else:
        raise TypeError("Please specify the rest analysis.")

    # subtract baseline offset
    signal_zero_mean = signal - np.mean(signal)

    # full-wave rectification
    fwlo = np.abs(signal_zero_mean)

    # moving average
    mvgav = np.convolve(fwlo, np.ones((size,)) / size, mode='valid')

    # calculate the test function
    tf = (1 / std_dev_rest) * (mvgav - mean_rest)

    onset_time_list = []
    offset_time_list = []
    alarm_time = 0
    state_duration = 0
    j = 0
    onset = False
    alarm = False
    for k in range(0, len(tf)):
        if onset is True:
            # an onset was previously detected and we are looking for the offset time applying the same criteria
            if alarm is False:  # if the alarm time has not yet been identified
                if tf[k] < threshold:  # alarm time
                    alarm_time = k
                    alarm = True
            else:  # now we have to check for the remaining rule to me bet - duration of inactive state
                if tf[k] < threshold:
                    state_duration += 1
                    if j > 0:  # there was one (or more) samples above the threshold level but now one is bellow it
                        # the test function may go above the threshold , but each time not longer than j samples
                        j = 0
                    if state_duration == active_state_duration:
                        offset_time_list.append(alarm_time)
                        onset = False
                        alarm = False
                        j = 0
                        state_duration = 0
                else:  # sample falls below the threshold level
                    j += 1
                    if j > fail_size:
                        # the inactive state is above the threshold for longer than the predefined number of samples
                        alarm = False
                        j = 0
                        state_duration = 0
        else:  # we only look for another onset if a previous offset was detected
            if alarm is False:  # if the alarm time has not yet been identified
                if tf[k] >= threshold:  # alarm time
                    alarm_time = k
                    alarm = True
            else:  # now we have to check for the remaining rule to me bet - duration of active state
                if tf[k] >= threshold:
                    state_duration += 1
                    if j > 0:  # there was one (or more) samples below the threshold level but now one is above it
                        # the test function may repeatedly fall below the threshold, but each time not longer than j samples
                        j = 0
                    if state_duration == active_state_duration:
                        onset_time_list.append(alarm_time)
                        onset = True
                        alarm = False
                        j = 0
                        state_duration = 0
                else:  # sample falls below the threshold level
                    j += 1
                    if j > fail_size:
                        # the active state has fallen below the threshold for longer than the predefined number of samples
                        alarm = False
                        j = 0
                        state_duration = 0

    onsets = np.union1d(onset_time_list,
                        offset_time_list)

    # adjust indices because of moving average
    onsets += int(size / 2)

    return utils.ReturnTuple((onsets, tf), ('onsets', 'processed'))


def abbink_onset_detector(signal=None, rest=None, sampling_rate=1000.,
                          size=None, alarm_size=None, threshold=None,
                          transition_threshold=None):
    """Determine onsets of EMG pulses.

    Follows the approach by Abbink et al.. [Abb98]_.

    Parameters
    ----------
    signal : array
        Input filtered EMG signal.
    rest : array, list, dict
        One of the following 3 options:
        * N-dimensional array with filtered samples corresponding to a
        rest period;
        * 2D array or list with the beginning and end indices of a segment of
        the signal corresponding to a rest period;
        * Dictionary with {'mean': mean value, 'std_dev': standard variation}.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    size : int
        Detection window size (seconds).
    alarm_size : int
        Number of amplitudes searched in the calculation of the transition
        index.
    threshold : int, float
        Detection threshold.
    transition_threshold: int, float
        Threshold used in the calculation of the transition index.

    Returns
    -------
    onsets : array
        Indices of EMG pulse onsets.
    processed : array
        Processed EMG signal.

    References
    ----------
    .. [Abb98] Abbink JH, van der Bilt A, van der Glas HW, "Detection of onset
       and termination of muscle activity in surface electromyograms",
       Journal of Oral Rehabilitation, vol. 25, pp. 365–369, 1998

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if rest is None:
        raise TypeError("Please specidy rest parameters.")

    if size is None:
        raise TypeError("Please specify the detection window size.")

    if alarm_size is None:
        raise TypeError("Please specify the number of amplitudes searched in "
                        "the calculation of the transition index.")

    if threshold is None:
        raise TypeError("Please specify the detection threshold.")

    if transition_threshold is None:
        raise TypeError("Please specify the second threshold.")

    # gather statistics on rest signal
    if isinstance(rest, np.ndarray) or isinstance(rest, list):
        # if the input parameter is a numpy array or a list
        if len(rest) >= 2:
            # first ensure numpy
            rest = np.array(rest)
            if len(rest) == 2:
                # the rest signal is a segment of the signal
                rest_signal = signal[rest[0]:rest[1]]
            else:
                # the rest signal is provided as is
                rest_signal = rest
            rest_zero_mean = rest_signal - np.mean(rest_signal)
            statistics = st.signal_stats(signal=rest_zero_mean)
            mean_rest = statistics['mean']
            std_dev_rest = statistics['std_dev']
        else:
            raise TypeError("Please specify the rest analysis.")
    elif isinstance(rest, dict):
        # if the input is a dictionary
        mean_rest = rest['mean']
        std_dev_rest = rest['std_dev']
    else:
        raise TypeError("Please specify the rest analysis.")

    # subtract baseline offset
    signal_zero_mean = signal - np.mean(signal)

    # full-wave rectification
    fwlo = np.abs(signal_zero_mean)

    # moving average
    mvgav = np.convolve(fwlo, np.ones((size,)) / size, mode='valid')

    # calculate the test function
    tf = (1 / std_dev_rest) * (mvgav - mean_rest)

    # additional filter
    filtered_tf, _, _ = st.filter_signal(signal=tf,
                                         ftype='butter',
                                         band='lowpass',
                                         order=10,
                                         frequency=30,
                                         sampling_rate=sampling_rate)
    # convert from numpy array to list to use list comprehensions
    filtered_tf = filtered_tf.tolist()

    onset_time_list = []
    offset_time_list = []
    alarm_time = 0
    onset = False
    alarm = False
    for k in range(0, len(tf)):
        if onset is True:
            # an onset was previously detected and we are looking for the offset time, applying the same criteria
            if alarm is False:
                if filtered_tf[k] < threshold:
                    # the first index of the sliding window is used as an estimate for the onset time (simple post-processor)
                    alarm_time = k
                    alarm = True
            else:
                # if alarm_time > alarm_window_size and len(emg_conditioned_list) == (alarm_time + alarm_window_size + 1):
                if alarm_time > alarm_size and k == (alarm_time + alarm_size + 1):
                    transition_indices = []
                    for j in range(alarm_size, alarm_time):
                        low_list = [filtered_tf[j-alarm_size+a] for a in range(1, alarm_size+1)]
                        low = sum(i < transition_threshold for i in low_list)
                        high_list = [filtered_tf[j+b] for b in range(1, alarm_size+1)]
                        high = sum(i > transition_threshold for i in high_list)
                        transition_indices.append(low + high)
                    offset_time_list = np.where(transition_indices == np.amin(transition_indices))[0].tolist()
                    onset = False
                    alarm = False
        else:  # we only look for another onset if a previous offset was detected
            if alarm is False:
                if filtered_tf[k] >= threshold:
                    # the first index of the sliding window is used as an estimate for the onset time (simple post-processor)
                    alarm_time = k
                    alarm = True
            else:
                # if alarm_time > alarm_window_size and len(emg_conditioned_list) == (alarm_time + alarm_window_size + 1):
                if alarm_time > alarm_size and k == (alarm_time + alarm_size + 1):
                    transition_indices = []
                    for j in range(alarm_size, alarm_time):
                        low_list = [filtered_tf[j-alarm_size+a] for a in range(1, alarm_size+1)]
                        low = sum(i < transition_threshold for i in low_list)
                        high_list = [filtered_tf[j+b] for b in range(1, alarm_size+1)]
                        high = sum(i > transition_threshold for i in high_list)
                        transition_indices.append(low + high)
                    onset_time_list = np.where(transition_indices == np.amax(transition_indices))[0].tolist()
                    onset = True
                    alarm = False

    onsets = np.union1d(onset_time_list,
                        offset_time_list)

    # adjust indices because of moving average
    onsets += int(size / 2)

    return utils.ReturnTuple((onsets, filtered_tf), ('onsets', 'processed'))


def solnik_onset_detector(signal=None, rest=None, sampling_rate=1000.,
                          threshold=None, active_state_duration=None):
    """Determine onsets of EMG pulses.

    Follows the approach by Solnik et al. [Sol10]_.

    Parameters
    ----------
    signal : array
        Input filtered EMG signal.
    rest : array, list, dict
        One of the following 3 options:
        * N-dimensional array with filtered samples corresponding to a
        rest period;
        * 2D array or list with the beginning and end indices of a segment of
        the signal corresponding to a rest period;
        * Dictionary with {'mean': mean value, 'std_dev': standard variation}.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    threshold : int, float
        Scale factor for calculating the detection threshold.
    active_state_duration: int
        Minimum duration of the active state.

    Returns
    -------
    onsets : array
        Indices of EMG pulse onsets.
    processed : array
        Processed EMG signal.

    References
    ----------
    .. [Sol10] Solnik S, Rider P, Steinweg K, DeVita P, Hortobágyi T,
       "Teager-Kaiser energy operator signal conditioning improves EMG onset
       detection", European Journal of Applied Physiology, vol 110:3,
       pp. 489-498, 2010

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if rest is None:
        raise TypeError("Please specidy rest parameters.")

    if threshold is None:
        raise TypeError("Please specify the scale factor for calculating the "
                        "detection threshold.")

    if active_state_duration is None:
        raise TypeError("Please specify the mininum duration of the "
                        "active state.")

    # gather statistics on rest signal
    if isinstance(rest, np.ndarray) or isinstance(rest, list):
        # if the input parameter is a numpy array or a list
        if len(rest) >= 2:
            # first ensure numpy
            rest = np.array(rest)
            if len(rest) == 2:
                # the rest signal is a segment of the signal
                rest_signal = signal[rest[0]:rest[1]]
            else:
                # the rest signal is provided as is
                rest_signal = rest
            rest_zero_mean = rest_signal - np.mean(rest_signal)
            statistics = st.signal_stats(signal=rest_zero_mean)
            mean_rest = statistics['mean']
            std_dev_rest = statistics['std_dev']
        else:
            raise TypeError("Please specify the rest analysis.")
    elif isinstance(rest, dict):
        # if the input is a dictionary
        mean_rest = rest['mean']
        std_dev_rest = rest['std_dev']
    else:
        raise TypeError("Please specify the rest analysis.")

    # subtract baseline offset
    signal_zero_mean = signal - np.mean(signal)

    # calculate threshold
    threshold = mean_rest + threshold * std_dev_rest

    tf_list = []
    onset_time_list = []
    offset_time_list = []
    alarm_time = 0
    state_duration = 0
    onset = False
    alarm = False
    for k in range(1, len(signal_zero_mean)-1):
        # calculate the test function
        # Teager-Kaiser energy operator
        tf = signal_zero_mean[k]**2 - signal_zero_mean[k+1] * signal_zero_mean[k-1]
        # full-wave rectification
        tf = np.abs(tf)
        tf_list.append(tf)
        if onset is True:
            # an onset was previously detected and we are looking for the offset time, applying the same criteria
            if alarm is False:  # if the alarm time has not yet been identified
                if tf < threshold:  # alarm time
                    alarm_time = k
                    alarm = True
            else:  # now we have to check for the remaining rule to me bet - duration of inactive state
                if tf < threshold:
                    state_duration += 1
                    if state_duration == active_state_duration:
                        offset_time_list.append(alarm_time)
                        onset = False
                        alarm = False
                        state_duration = 0
        else:  # we only look for another onset if a previous offset was detected
            if alarm is False:  # if the alarm time has not yet been identified
                if tf >= threshold:  # alarm time
                    alarm_time = k
                    alarm = True
            else:  # now we have to check for the remaining rule to me bet - duration of active state
                if tf >= threshold:
                    state_duration += 1
                    if state_duration == active_state_duration:
                        onset_time_list.append(alarm_time)
                        onset = True
                        alarm = False
                        state_duration = 0

    onsets = np.union1d(onset_time_list,
                        offset_time_list)

    return utils.ReturnTuple((onsets, tf_list), ('onsets', 'processed'))


def silva_onset_detector(signal=None, sampling_rate=1000.,
                         size=None, threshold_size=None, threshold=None):
    """Determine onsets of EMG pulses.

    Follows the approach by Silva et al. [Sil12]_.

    Parameters
    ----------
    signal : array
        Input filtered EMG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    size : int
        Detection window size (seconds).
    threshold_size : int
        Window size for calculation of the adaptive threshold; must be bigger
        than the detection window size.
    threshold : int, float
        Fixed threshold for the double criteria.

    Returns
    -------
    onsets : array
        Indices of EMG pulse onsets.
    processed : array
        Processed EMG signal.

    References
    ----------
    .. [Sil12] Silva H, Scherer R, Sousa J, Londral A , "Towards improving the
       usability of electromyographic interfacess", Journal of Oral
       Rehabilitation, pp. 1–2, 2012

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if size is None:
        raise TypeError("Please specify the detection window size.")

    if threshold_size is None:
        raise TypeError("Please specify the window size for calculation of "
                        "the adaptive threshold.")

    if threshold_size <= size:
        raise TypeError("The window size for calculation of the adaptive "
                        "threshold must be bigger than the detection "
                        "window size")

    if threshold is None:
        raise TypeError("Please specify the fixed threshold for the "
                        "double criteria.")

    # subtract baseline offset
    signal_zero_mean = signal - np.mean(signal)

    # full-wave rectification
    fwlo = np.abs(signal_zero_mean)

    # moving average for calculating the test function
    tf_mvgav = np.convolve(fwlo, np.ones((size,)) / size, mode='valid')

    # moving average for calculating the adaptive threshold
    threshold_mvgav = np.convolve(fwlo, np.ones((threshold_size,)) / threshold_size, mode='valid')

    onset_time_list = []
    offset_time_list = []
    onset = False
    for k in range(0, len(threshold_mvgav)):
        if onset is True:
            # an onset was previously detected and we are looking for the offset time, applying the same criteria
            if tf_mvgav[k] < threshold_mvgav[k] and tf_mvgav[k] < threshold:
                offset_time_list.append(k)
                onset = False  # the offset has been detected, and we can look for another activation
        else:  # we only look for another onset if a previous offset was detected
            if tf_mvgav[k] >= threshold_mvgav[k] and tf_mvgav[k] >= threshold:
                # the first index of the sliding window is used as an estimate for the onset time (simple post-processor)
                onset_time_list.append(k)
                onset = True

    onsets = np.union1d(onset_time_list,
                        offset_time_list)

    # adjust indices because of moving average
    onsets += int(size / 2)

    return utils.ReturnTuple((onsets, tf_mvgav), ('onsets', 'processed'))


def londral_onset_detector(signal=None, rest=None, sampling_rate=1000.,
                           size=None, threshold=None,
                           active_state_duration=None):
    """Determine onsets of EMG pulses.

    Follows the approach by Londral et al. [Lon13]_.

    Parameters
    ----------
    signal : array
        Input filtered EMG signal.
    rest : array, list, dict
        One of the following 3 options:
        * N-dimensional array with filtered samples corresponding to a
        rest period;
        * 2D array or list with the beginning and end indices of a segment of
        the signal corresponding to a rest period;
        * Dictionary with {'mean': mean value, 'std_dev': standard variation}.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    size : int
        Detection window size (seconds).
    threshold : int, float
        Scale factor for calculating the detection threshold.
    active_state_duration: int
        Minimum duration of the active state.

    Returns
    -------
    onsets : array
        Indices of EMG pulse onsets.
    processed : array
        Processed EMG signal.

    References
    ----------
    .. [Lon13] Londral A, Silva H, Nunes N, Carvalho M, Azevedo L, "A wireless
       user-computer interface to explore various sources of biosignals and
       visual biofeedback for severe motor impairment",
       Journal of Accessibility and Design for All, vol. 3:2, pp. 118–134, 2013
    
    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if rest is None:
        raise TypeError("Please specidy rest parameters.")

    if size is None:
        raise TypeError("Please specify the detection window size.")

    if threshold is None:
        raise TypeError("Please specify the scale factor for calculating the "
                        "detection threshold.")

    if active_state_duration is None:
        raise TypeError("Please specify the mininum duration of the "
                        "active state.")

    # gather statistics on rest signal
    if isinstance(rest, np.ndarray) or isinstance(rest, list):
        # if the input parameter is a numpy array or a list
        if len(rest) >= 2:
            # first ensure numpy
            rest = np.array(rest)
            if len(rest) == 2:
                # the rest signal is a segment of the signal
                rest_signal = signal[rest[0]:rest[1]]
            else:
                # the rest signal is provided as is
                rest_signal = rest
            rest_zero_mean = rest_signal - np.mean(rest_signal)
            statistics = st.signal_stats(signal=rest_zero_mean)
            mean_rest = statistics['mean']
            std_dev_rest = statistics['std_dev']
        else:
            raise TypeError("Please specify the rest analysis.")
    elif isinstance(rest, dict):
        # if the input is a dictionary
        mean_rest = rest['mean']
        std_dev_rest = rest['std_dev']
    else:
        raise TypeError("Please specify the rest analysis.")

    # subtract baseline offset
    signal_zero_mean = signal - np.mean(signal)

    # calculate threshold
    threshold = mean_rest + threshold * std_dev_rest

    # helper function for calculating the test function for each window
    def _londral_test_function(signal=None):
        tf = (1 / size) * (sum(j ** 2 for j in signal) - (1 / size) * (sum(signal) ** 2))
        return tf

    # calculate the test function
    _, tf = st.windower(
        signal=signal_zero_mean,
        size=size, step=1,
        fcn=_londral_test_function,
        kernel='rectangular',
    )

    onset_time_list = []
    offset_time_list = []
    alarm_time = 0
    state_duration = 0
    onset = False
    alarm = False
    for k in range(0, len(tf)):
        if onset is True:
            # an onset was previously detected and we are looking for the offset time, applying the same criteria
            if alarm is False:  # if the alarm time has not yet been identified
                if tf[k] < threshold:  # alarm time
                    alarm_time = k
                    alarm = True
            else:  # now we have to check for the remaining rule to me bet - duration of inactive state
                if tf[k] < threshold:
                    state_duration += 1
                    if state_duration == active_state_duration:
                        offset_time_list.append(alarm_time)
                        onset = False
                        alarm = False
                        state_duration = 0
        else:  # we only look for another onset if a previous offset was detected
            if alarm is False:  # if the alarm time has not yet been identified
                if tf[k] >= threshold:  # alarm time
                    alarm_time = k
                    alarm = True
            else:  # now we have to check for the remaining rule to me bet - duration of active state
                if tf[k] >= threshold:
                    state_duration += 1
                    if state_duration == active_state_duration:
                        onset_time_list.append(alarm_time)
                        onset = True
                        alarm = False
                        state_duration = 0

    onsets = np.union1d(onset_time_list,
                        offset_time_list)

    # adjust indices because of moving average
    onsets += int(size / 2)

    return utils.ReturnTuple((onsets, tf), ('onsets', 'processed'))



# -*- coding: utf-8 -*-
"""
biosppy.signals.ppg
-------------------

This module provides methods to process Photoplethysmogram (PPG) signals.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
from six.moves import range

# 3rd party
import numpy as np
import scipy.signal as ss
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde

# local
from . import tools as st
from .. import plotting, utils


def ppg(signal=None, sampling_rate=1000., show=True):
    """Process a raw PPG signal and extract relevant signal features using
    default parameters.

    Parameters
    ----------
    signal : array
        Raw PPG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    show : bool, optional
        If True, show a summary plot.

    Returns
    -------
    ts : array
        Signal time axis reference (seconds).
    filtered : array
        Filtered PPG signal.
    onsets : array
        Indices of PPG pulse onsets.
    heart_rate_ts : array
        Heart rate time axis reference (seconds).
    heart_rate : array
        Instantaneous heart rate (bpm).

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    sampling_rate = float(sampling_rate)

    # filter signal
    filtered, _, _ = st.filter_signal(signal=signal,
                                      ftype='butter',
                                      band='bandpass',
                                      order=4,
                                      frequency=[1, 8],
                                      sampling_rate=sampling_rate)

    # find onsets
    onsets, _ = find_onsets_elgendi2013(signal=filtered, sampling_rate=sampling_rate)

    # compute heart rate
    hr_idx, hr = st.get_heart_rate(beats=onsets,
                                   sampling_rate=sampling_rate,
                                   smooth=True,
                                   size=3)

    # get time vectors
    length = len(signal)
    T = (length - 1) / sampling_rate
    ts = np.linspace(0, T, length, endpoint=False)
    ts_hr = ts[hr_idx]

    # plot
    if show:
        plotting.plot_ppg(ts=ts,
                          raw=signal,
                          filtered=filtered,
                          onsets=onsets,
                          heart_rate_ts=ts_hr,
                          heart_rate=hr,
                          path=None,
                          show=True)

    # output
    args = (ts, filtered, onsets, ts_hr, hr)
    names = ('ts', 'filtered', 'onsets', 'heart_rate_ts', 'heart_rate')

    return utils.ReturnTuple(args, names)

def find_onsets_elgendi2013(signal=None, sampling_rate=1000., peakwindow=0.111, beatwindow=0.667, beatoffset=0.02, mindelay=0.3):
    """
    Determines onsets of PPG pulses.

    Parameters
    ----------
    signal : array
        Input filtered PPG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    peakwindow : float
        Parameter W1 on referenced article
        Optimized at 0.111
    beatwindow : float
        Parameter W2 on referenced article
        Optimized at 0.667
    beatoffset : float
        Parameter beta on referenced article
        Optimized at 0.2
    mindelay : float
        Minimum delay between peaks.
        Avoids false positives

    Returns
    ----------
    onsets : array
        Indices of PPG pulse onsets.
    params : dict
        Input parameters of the function


    References
    ----------
    - Elgendi M, Norton I, Brearley M, Abbott D, Schuurmans D (2013) Systolic Peak Detection in
    Acceleration Photoplethysmograms Measured from Emergency Responders in Tropical Conditions.
    PLoS ONE 8(10): e76585. doi:10.1371/journal.pone.0076585.
    
    Notes
    ---------------------
    Optimal ranges for signal filtering (from Elgendi et al. 2013):
    "Optimization of the beat detector’s spectral window for the lower frequency resulted in a 
    value within 0.5– 1 Hz with the higher frequency within 7–15 Hz"
    
    All the number references below between curly brackets {...} by the code refer to the line numbers of
    code in "Table 2 Algorithm IV: DETECTOR (PPG signal, F1, F2, W1, W2, b)" from Elgendi et al. 2013 for a
    better comparison of the algorithm
    
    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # Create copy of signal (not to modify the original object)
    signal_copy = np.copy(signal)

    # Truncate to zero and square
    # {3, 4}
    signal_copy[signal_copy < 0] = 0
    squared_signal = signal_copy ** 2

    # Calculate peak detection threshold
    # {5}
    ma_peak_kernel = int(np.rint(peakwindow * sampling_rate))
    ma_peak, _ = st.smoother(squared_signal, kernel="boxcar", size=ma_peak_kernel)

    # {6}
    ma_beat_kernel = int(np.rint(beatwindow * sampling_rate))
    ma_beat, _ = st.smoother(squared_signal, kernel="boxcar", size=ma_beat_kernel)

    # Calculate threshold value
    # {7, 8, 9}
    thr1 = ma_beat + beatoffset * np.mean(squared_signal)

    # Identify start and end of PPG waves.
    # {10-16}
    waves = ma_peak > thr1
    beg_waves = np.where(np.logical_and(np.logical_not(waves[0:-1]), waves[1:]))[0]
    end_waves = np.where(np.logical_and(waves[0:-1], np.logical_not(waves[1:])))[0]
    # Throw out wave-ends that precede first wave-start.
    end_waves = end_waves[end_waves > beg_waves[0]]

    # Identify systolic peaks within waves (ignore waves that are too short).
    num_waves = min(beg_waves.size, end_waves.size)
    # {18}
    min_len = int(np.rint(peakwindow * sampling_rate))
    min_delay = int(np.rint(mindelay * sampling_rate))
    onsets = [0]

    # {19}
    for i in range(num_waves):

        beg = beg_waves[i]
        end = end_waves[i]
        len_wave = end - beg

        # {20, 22, 23}
        if len_wave < min_len:
            continue

        # Find local maxima and their prominence within wave span.
        # {21}
        data = signal_copy[beg:end]
        locmax, props = ss.find_peaks(data, prominence=(None, None))

        # If more than one peak
        if locmax.size > 0:
            # Identify most prominent local maximum.
            peak = beg + locmax[np.argmax(props["prominences"])]
            # Enforce minimum delay between onsets.
            if peak - onsets[-1] > min_delay:
                onsets.append(peak)

    onsets.pop(0)
    onsets = np.array(onsets, dtype='int')

    # output
    params = {'signal': signal, 'sampling_rate': sampling_rate, 'peakwindow': peakwindow, 'beatwindow': beatwindow, 'beatoffset': beatoffset, 'mindelay': mindelay}

    args = (onsets, params)
    names = ('onsets', 'params')

    return utils.ReturnTuple(args, names)


def find_onsets_kavsaoglu2016(
    signal=None,
    sampling_rate=1000.0,
    alpha=0.2,
    k=4,
    init_bpm=90,
    min_delay=0.6,
    max_BPM=150,
):
    """
    Determines onsets of PPG pulses.

    Parameters
    ----------
    signal : array
        Input filtered PPG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    alpha : float, optional
        Low-pass filter factor.
        Avoids abrupt changes of BPM.
    k : int, float, optional
        Number of segments by pulse.
        Width of each segment = Period of pulse according to current BPM / k
    init_bpm : int, float, optional
        Initial BPM.
        Higher value results in a smaller segment width.
    min_delay : float
        Minimum delay between peaks as percentage of current BPM pulse period.
        Avoids false positives
    max_bpm : int, float, optional
        Maximum BPM.
        Maximum value accepted as valid BPM.

    Returns
    ----------
    onsets : array
        Indices of PPG pulse onsets.
    window_marks : array
        Indices of segments window boundaries.
    params : dict
        Input parameters of the function


    References
    ----------
    - Kavsaoğlu, Ahmet & Polat, Kemal & Bozkurt, Mehmet. (2016). An innovative peak detection algorithm for
    photoplethysmography signals: An adaptive segmentation method. TURKISH JOURNAL OF ELECTRICAL ENGINEERING
    & COMPUTER SCIENCES. 24. 1782-1796. 10.3906/elk-1310-177.

    Notes
    ---------------------
    This algorithm is an adaption of the one described on Kavsaoğlu et al. (2016).
    This version takes into account a minimum delay between peaks and builds upon the adaptive segmentation
    by using a low-pass filter for BPM changes. This way, even if the algorithm wrongly detects a peak, the
    BPM value will stay relatively constant so the next pulse can be correctly segmented.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if alpha <= 0 or alpha > 1:
        raise TypeError("The value of alpha must be in the range: ]0, 1].")

    if k <= 0:
        raise TypeError("The number of divisions by pulse should be greater than 0.")

    if init_bpm <= 0:
        raise TypeError("Provide a valid BPM value for initial estimation.")

    if min_delay < 0 or min_delay > 1:
        raise TypeError(
            "The minimum delay percentage between peaks must be between 0 and 1"
        )

    if max_BPM >= 248:
        raise TypeError("The maximum BPM must assure the person is alive")

    # current bpm
    bpm = init_bpm

    # current segment window width
    window = int(sampling_rate * (60 / bpm) / k)

    # onsets array
    onsets = []

    # window marks array - stores the boundaries of each segment
    window_marks = []

    # buffer for peak indices
    idx_buffer = [-1, -1, -1]

    # buffer to store the previous 3 values for onset detection
    min_buffer = [0, 0, 0]

    # signal pointer
    i = 0
    while i + window < len(signal):
        # remove oldest values
        idx_buffer.pop(0)
        min_buffer.pop(0)

        # add the index of the minimum value of the current segment to buffer
        idx_buffer.append(int(i + np.argmin(signal[i : i + window])))

        # add the minimum value of the current segment to buffer
        min_buffer.append(signal[idx_buffer[-1]])

        if (
            # the buffer has to be filled with valid values
            idx_buffer[0] > -1
            # the center value of the buffer must be smaller than its neighbours
            and (min_buffer[1] < min_buffer[0] and min_buffer[1] <= min_buffer[2])
            # if an onset was previously detected, guarantee that the new onset respects the minimum delay, minimum BPM and maximum BPM
            and (
                len(onsets) == 0
                or (
                    (idx_buffer[1] - onsets[-1]) / sampling_rate >= min_delay * 60 / bpm
                    and (idx_buffer[1] - onsets[-1]) / sampling_rate > 60 / max_BPM
                )
            )
        ):
            # store the onset
            onsets.append(idx_buffer[1])

            # if more than one onset was detected, update the bpm and the segment width
            if len(onsets) > 1:
                # calculate new bpm from the latest two onsets
                new_bpm = int(60 * sampling_rate / (onsets[-1] - onsets[-2]))

                # update the bpm value
                bpm = alpha * new_bpm + (1 - alpha) * bpm

                # update the segment window width
                window = int(sampling_rate * (60 / bpm) / k)

        # update the signal pointer
        i += window

        # store window segment boundaries index
        window_marks.append(i)

    onsets = np.array(onsets, dtype="int")
    window_marks = np.array(window_marks, dtype="int")

    # output
    params = {
        "signal": signal,
        "sampling_rate": sampling_rate,
        "alpha": alpha,
        "k": k,
        "init_bpm": init_bpm,
        "min_delay": min_delay,
        "max_bpm": max_BPM,
    }

    args = (onsets, window_marks, params)
    names = ("onsets", "window_marks", "params")

    return utils.ReturnTuple(args, names)


def ppg_segmentation(filtered,
                     sampling_rate=1000.,
                     show=False,
                     show_mean=False,
                     selection=False,
                     peak_threshold=None):
    """"Segments a filtered PPG signal. Segmentation filtering is achieved by
    taking into account segments selected by peak height and pulse morphology.

    Parameters
    ----------
    filtered : array
        Filtered PPG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    show : bool, optional
        If True, show a plot with segments. Segments are clipped.
    show_mean : bool, optional
        If True, shows the mean pulse on top of segments.
    selection : bool, optional
        If True, performs selection with peak height and pulse morphology.
    peak_threshold : int, float, optional
        If `selection` is True, selects peaks with height greater than defined
        threshold.

    Returns
    -------
    segments : array
        Start and end indices for each detected pulse segment.
    selected_segments : array
        Start and end indices for each selected pulse segment.
    mean_pulse_ts : array
        Mean pulse time axis reference (seconds).
    mean_pulse : array
        Mean wave of clipped PPG pulses.
    onsets : array
        Indices of PPG pulse onsets. Onsets are found based on minima.
    peaks : array
        Indices of PPG pulse peaks. 'Elgendi2013' algorithm is used.

    """

    # check inputs
    if filtered is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    filtered = np.array(filtered)

    sampling_rate = float(sampling_rate)

    # find peaks (last peak is rejected)
    peaks, _ = find_onsets_elgendi2013(filtered)
    nb_segments = len(peaks) - 1

    # find minima
    minima = (np.diff(np.sign(np.diff(filtered))) > 0).nonzero()[0]

    # find onsets
    onsets = []

    for i in peaks:
        # find the lower closest number to the peak
        onsets.append(minima[minima < i].max())

    onsets = np.array(onsets, dtype='int')

    if len(peaks) == 0 or len(onsets) == 0:
        raise TypeError("No peaks or onsets detected.")

    # define peak threshold with peak density function (for segment selection),
    # where the maximum value is choosen.
    if peak_threshold == None and selection:
        density = gaussian_kde(filtered[peaks])
        xs = np.linspace(0, max(filtered[peaks]), 1000)
        density.covariance_factor = lambda : .25
        density._compute_covariance()
        peak_threshold = xs[np.argmax(density(xs))]

    # segments array with start and end indexes, and segment selection
    segments = np.zeros((nb_segments, 2), dtype='int')
    segments_sel = []

    for i in range(nb_segments):

        # assign start and end of each segment
        segments[i, 0] = onsets[i]
        segments[i, 1] = onsets[i + 1]

        # search segments with at least 4 max+min (standard waveform) and
        # peak height greater than threshold for pulse selection
        if selection:
            seg = filtered[segments[i, 0] : segments[i, 1]]
            if max(seg) > peak_threshold:
                if len(np.where(np.diff(np.sign(np.diff(seg))))[0]) > 3:
                    segments_sel.append(i)

            if len(segments_sel) == 0 :
                print('Warning: Suitable waves not found. [-0.1, 0.4]s cut from peak is made.')

    # find earliest onset-peak duration (ensure minimal shift of 0.1s)
    shifts = peaks - onsets

    cut1 = 0.1*sampling_rate
    if len(segments_sel) > 0 and selection:
        shifts_sel = np.take(shifts, segments_sel)
        shifts_sel = shifts_sel[shifts_sel > 0.1*sampling_rate]
        cut1 = min(shifts_sel)

    # find shortest peak-end duration (ensure minimal duration of 0.4s)
    cut2 = 0.4*sampling_rate
    ep_d = segments[:, 1] - peaks[0 : len(segments)]
    if len(segments_sel) > 0 and selection:
        ep_d_sel = np.take(ep_d, segments_sel)
        ep_d_sel = ep_d_sel[ep_d_sel > 0.4*sampling_rate]
        cut2 = min(ep_d_sel)

    # clipping segments
    c_segments = np.zeros((nb_segments, 2), dtype=int)
    for i in range(nb_segments):
        c_segments[i, 0] = peaks[i] - cut1
        c_segments[i, 1] = peaks[i] + cut2

    cut_length = c_segments[0, 1] - c_segments[0, 0]


    # time axis
    mean_pulse_ts = np.arange(0, cut_length/sampling_rate, 1./sampling_rate)

    # plot
    if show:
        # figure layout
        fig, ax = plt.subplots()
        fig.suptitle('PPG Segments', fontweight='bold', x=0.51)
        ax.set_xlabel('Time (s)')
        ax.set_ylabel('Amplitude (a.u.)')

    # sum of segments to plot mean wave pulse
    sum_segments = np.zeros(cut_length)

    # transparency factor to plot segments (alpha)
    b = 1 - np.log(1. - 0.01)
    alpha = np.exp(-nb_segments + b) + 0.01
    if alpha > 1:
        alpha = 1

    # plot segments
    if selection:
        for i in segments_sel:
            wave = filtered[c_segments[i, 0] : c_segments[i, 1]]
            if show:
                ax.plot(mean_pulse_ts, wave, color='tab:blue', alpha=alpha)
            sum_segments = sum_segments + wave
            ax.set_title(f'[selection only, {len(segments_sel)} segment(s)]')

    else:
        for i in range(nb_segments):
            wave = filtered[c_segments[i, 0] : c_segments[i, 1]]
            if show:
                ax.plot(mean_pulse_ts, wave, color='tab:blue', alpha=alpha)
                ax.set_title(f'[{nb_segments} segment(s)]')
            sum_segments = sum_segments + wave

    # plot mean pulse
    mean_pulse = sum_segments/len(segments_sel)
    if show and show_mean:
        ax.plot(mean_pulse_ts, mean_pulse, color='tab:orange', label='Mean wave')
        ax.legend()

    # output
    args = (segments, segments[segments_sel], mean_pulse_ts, mean_pulse, onsets, peaks)
    names = ('segments', 'selected_segments', 'mean_pulse_ts', 'mean_pulse', 'onsets', 'peaks')

    return utils.ReturnTuple(args, names)



# -*- coding: utf-8 -*-
"""
biosppy.signals.eda
-------------------

This module provides methods to process Electrodermal Activity (EDA)
signals, also known as Galvanic Skin Response (GSR).

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
from six.moves import range

# 3rd party
import numpy as np

# local
from . import tools as st
from .. import plotting, utils


def eda(signal=None, sampling_rate=1000.0, path=None, show=True, min_amplitude=0.1):
    """Process a raw EDA signal and extract relevant signal features using
    default parameters.

    Parameters
    ----------
    signal : array
        Raw EDA signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show a summary plot.
    min_amplitude : float, optional
        Minimum treshold by which to exclude SCRs.

    Returns
    -------
    ts : array
        Signal time axis reference (seconds).
    filtered : array
        Filtered EDA signal.
    onsets : array
        Indices of SCR pulse onsets.
    peaks : array
        Indices of the SCR peaks.
    amplitudes : array
        SCR pulse amplitudes.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    sampling_rate = float(sampling_rate)

    # filter signal
    aux, _, _ = st.filter_signal(
        signal=signal,
        ftype="butter",
        band="lowpass",
        order=4,
        frequency=5,
        sampling_rate=sampling_rate,
    )

    # smooth
    sm_size = int(0.75 * sampling_rate)
    filtered, _ = st.smoother(signal=aux, kernel="boxzen", size=sm_size, mirror=True)

    # get SCR info
    onsets, peaks, amps = kbk_scr(
        signal=filtered, sampling_rate=sampling_rate, min_amplitude=min_amplitude
    )

    # get time vectors
    length = len(signal)
    T = (length - 1) / sampling_rate
    ts = np.linspace(0, T, length, endpoint=True)

    # plot
    if show:
        plotting.plot_eda(
            ts=ts,
            raw=signal,
            filtered=filtered,
            onsets=onsets,
            peaks=peaks,
            amplitudes=amps,
            path=path,
            show=True,
        )

    # output
    args = (ts, filtered, onsets, peaks, amps)
    names = ("ts", "filtered", "onsets", "peaks", "amplitudes")

    return utils.ReturnTuple(args, names)


def basic_scr(signal=None, sampling_rate=1000.0):
    """Basic method to extract Skin Conductivity Responses (SCR) from an
    EDA signal.

    Follows the approach in [Gamb08]_.

    Parameters
    ----------
    signal : array
        Input filterd EDA signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).

    Returns
    -------
    onsets : array
        Indices of the SCR onsets.
    peaks : array
        Indices of the SRC peaks.
    amplitudes : array
        SCR pulse amplitudes.

    References
    ----------
    .. [Gamb08] Hugo Gamboa, "Multi-modal Behavioral Biometrics Based on HCI
       and Electrophysiology", PhD thesis, Instituto Superior T{\'e}cnico, 2008

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # find extrema
    pi, _ = st.find_extrema(signal=signal, mode="max")
    ni, _ = st.find_extrema(signal=signal, mode="min")

    # sanity check
    if len(pi) == 0 or len(ni) == 0:
        raise ValueError("Could not find SCR pulses.")

    # pair vectors
    if ni[0] > pi[0]:
        ni = ni[1:]
    if pi[-1] < ni[-1]:
        pi = pi[:-1]
    if len(pi) > len(ni):
        pi = pi[:-1]

    li = min(len(pi), len(ni))
    i1 = pi[:li]
    i3 = ni[:li]

    # indices
    i0 = np.array((i1 + i3) / 2.0, dtype=int)
    if i0[0] < 0:
        i0[0] = 0

    # amplitude
    a = signal[i0] - signal[i3]

    # output
    args = (i3, i0, a)
    names = ("onsets", "peaks", "amplitudes")

    return utils.ReturnTuple(args, names)


def kbk_scr(signal=None, sampling_rate=1000.0, min_amplitude=0.1):
    """KBK method to extract Skin Conductivity Responses (SCR) from an
    EDA signal.

    Follows the approach by Kim *et al.* [KiBK04]_.

    Parameters
    ----------
    signal : array
        Input filterd EDA signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    min_amplitude : float, optional
        Minimum treshold by which to exclude SCRs.

    Returns
    -------
    onsets : array
        Indices of the SCR onsets.
    peaks : array
        Indices of the SRC peaks.
    amplitudes : array
        SCR pulse amplitudes.

    References
    ----------
    .. [KiBK04] K.H. Kim, S.W. Bang, and S.R. Kim, "Emotion recognition
       system using short-term monitoring of physiological signals",
       Med. Biol. Eng. Comput., vol. 42, pp. 419-427, 2004

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # differentiation
    df = np.diff(signal)

    # smooth
    size = int(1.0 * sampling_rate)
    df, _ = st.smoother(signal=df, kernel="bartlett", size=size, mirror=True)

    # zero crosses
    (zeros,) = st.zero_cross(signal=df, detrend=False)
    if np.all(df[: zeros[0]] > 0):
        zeros = zeros[1:]
    if np.all(df[zeros[-1] :] > 0):
        zeros = zeros[:-1]

    scrs, amps, ZC, pks = [], [], [], []
    for i in range(0, len(zeros) - 1, 2):
        scrs += [df[zeros[i] : zeros[i + 1]]]
        ZC += [zeros[i]]
        ZC += [zeros[i + 1]]
        pks += [zeros[i] + np.argmax(df[zeros[i] : zeros[i + 1]])]
        amps += [signal[pks[-1]] - signal[ZC[-2]]]

    # exclude SCRs with small amplitude
    thr = min_amplitude * np.max(amps)
    idx = np.where(amps > thr)

    scrs = np.array(scrs, dtype=np.object)[idx]
    amps = np.array(amps)[idx]
    ZC = np.array(ZC)[np.array(idx) * 2]
    pks = np.array(pks, dtype=int)[idx]

    onsets = ZC[0].astype(int)

    # output
    args = (onsets, pks, amps)
    names = ("onsets", "peaks", "amplitudes")

    return utils.ReturnTuple(args, names)



# -*- coding: utf-8 -*-
"""
biosppy.signals.ecg
-------------------

This module provides methods to synthesize Electrocardiographic (ECG) signals.

:copyright: (c) 2015-2021 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.

"""

# Imports
from math import pi

# 3rd party
import numpy as np
import biosppy.signals
import warnings
import matplotlib.pyplot as plt

# local
from biosppy.signals import tools as st
from .. import plotting, utils


def B(l, Kb):
    """Generates the amplitude values of the first isoelectric line (B segment) of the ECG signal.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameter introduced doesn't make sense in this context, an error will raise.
    Parameters
    ----------
    l  : float
        Inverse of the sampling rate.
    Kb : int
        B segment width (miliseconds).
    Returns
    -------
    B_segment : array
        B segment amplitude values (milivolts).

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    """
    if Kb > 130:
        raise Exception("Warning! Kb is out of boundaries.")
    else:
        a = np.zeros(Kb * l)
        B_segment = a.tolist()
    return B_segment


def P(i, Ap, Kp):
    """Generates the amplitude values of the P wave in the ECG signal.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameters introduced don't make sense in this context, an error will raise.
    Parameters
    ----------
    i  : int
        Sampling rate.
    Ap : int
        P wave amplitude (milivolts).
    Kp : int
        P wave width (miliseconds).
    Returns
    -------
    P_wave : array
        P wave amplitude values (milivolts).

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    """
    if Ap < -0.2 or Ap > 0.5:
        raise Exception("Warning! Ap is out of boundaries.")
    elif Kp < 10 or Kp > 100:
        raise Exception("Warning! Kp is out of boundaries.")
    else:
        k = np.arange(0, Kp, i)
        a = -(Ap / 2.0) * np.cos((2 * np.pi * k + 15) / Kp) + Ap / 2.0
        P_wave = a.tolist()
    return P_wave


def Pq(l, Kpq):
    """Generates the amplitude values of the PQ segment in the ECG signal.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameters introduced don't make sense in this context, an error will raise.
    Parameters
    ----------
    l  : float
        Inverse of the sampling rate.
    Kpq : int
        PQ segment width (miliseconds).
    Returns
    -------
    PQ_segment : array
        PQ segment amplitude values (milivolts).

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    """
    if Kpq < 0 or Kpq > 60:
        raise Exception("Warning! Kpq is out of boundaries.")
    else:
        a = np.zeros(Kpq * l)
        PQ_segment = a.tolist()
    return PQ_segment


def Q1(i, Aq, Kq1):
    """Generates the amplitude values of the first 5/6 of the Q wave in the ECG signal.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameters introduced don't make sense in this context, an error will raise.
    Parameters
    ----------
    i   : int
        Sampling rate.
    Aq  : int
        Q wave amplitude (milivolts).
    Kq1 : int
        First 5/6 of the Q wave width (miliseconds).
    Returns
    -------
    Q1_wave : array
        First 5/6 of the Q wave amplitude values (milivolts).

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    """
    if Aq < 0 or Aq > 0.5:
        raise Exception("Warning! Aq is out of boundaries.")
    elif Kq1 < 0 or Kq1 > 70:
        raise Exception("Warning! Kq1 is out of boundaries.")
    else:
        k = np.arange(0, Kq1, i)
        a = -Aq * (k / Kq1)
        Q1_wave = a.tolist()
    return Q1_wave


def Q2(i, Aq, Kq2):
    """Generates the amplitude values of the last 1/6 of the Q wave in the ECG signal.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameters introduced don't make sense in this context, an error will raise.
    Parameters
    ----------
    i   : int
        Sampling rate.
    Aq  : int
        Q wave amplitude (milivolts).
    Kq2 : int
        Last 1/6 of the Q wave width (miliseconds).
    Returns
    -------
    Q2_wave : array
        Last 1/6 of the Q wave amplitude values (milivolts).

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    """
    if Aq < 0 or Aq > 0.5:
        raise Exception("Warning! Aq is out of boundaries.")
    elif Kq2 < 0 or Kq2 > 50:
        raise Exception("Warning! Kq2 is out of boundaries.")
    else:
        k = np.arange(0, Kq2, i)
        a = Aq * (k / Kq2) - Aq
        Q2_wave = a.tolist()
    return Q2_wave


def R(i, Ar, Kr):
    """Generates the amplitude values of the R wave in the ECG signal.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameters introduced don't make sense in this context, an error will raise.
    Parameters
    ----------
    i  : int
        Sampling rate.
    Ar : int
        R wave amplitude (milivolts).
    Kr : int
        R wave width (miliseconds).
    Returns
    -------
    R_wave : array
        R wave amplitude values (milivolts).

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    """
    if Ar < 0.5 or Ar > 2:
        raise Exception("Warning! Ar is out of boundaries.")
    elif Kr < 10 or Kr > 150:
        raise Exception("Warning! Kr is out of boundaries.")
    else:
        k = np.arange(0, Kr, i)
        a = Ar * np.sin((np.pi * k) / Kr)
        R_wave = a.tolist()
    return R_wave


def S(i, As, Ks, Kcs, k=0):
    """Generates the amplitude values of the S wave in the ECG signal.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameters introduced don't make sense in this context, an error will raise.
    Parameters
    ----------
    i  : int
        Sampling rate.
    As : int
        S wave amplitude (milivolts).
    Ks : int
        S wave width (miliseconds).
    Kcs : int
        Parameter which allows slight adjustment of S wave shape by cutting away a portion at the end.
    k : int, optional
    Returns
    -------
    S : array
        If k = 0, S wave amplitude values (milivolts).
    S : int
        If k != 0, value obtained by using the S wave expression for the given k value.

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    """
    if As < 0 or As > 1:
        raise Exception("Warning! As is out of boundaries.")
    elif Ks < 10 or Ks > 200:
        raise Exception("Warning! Ks is out of boundaries.")
    elif Kcs < -5 or Kcs > 150:
        raise Exception("Warning! Kcs is out of boundaries.")
    else:
        if k == 0:
            k = np.arange(0, Ks - Kcs, i)
            a = (
                -As
                * i
                * k
                * (19.78 * np.pi)
                / Ks
                * np.exp(-2 * (((6 * np.pi) / Ks) * i * k) ** 2)
            )
            S = a.tolist()
        else:
            S = (
                -As
                * i
                * k
                * (19.78 * np.pi)
                / Ks
                * np.exp(-2 * (((6 * np.pi) / Ks) * i * k) ** 2)
            )
    return S


def St(i, As, Ks, Kcs, sm, Kst, k=0):
    """Generates the amplitude values of the ST segment in the ECG signal.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameters introduced don't make sense in this context, an error will raise.
    Parameters
    ----------
    i  : int
        Sampling rate.
    As : int
        S wave amplitude (milivolts).
    Ks : int
        S wave width (miliseconds).
    Kcs : int
        Parameter which allows slight adjustment of S wave shape by cutting away a portion at the end.
    sm : int
        Slope parameter in the ST segment.
    Kst : int
        ST segment width (miliseconds).
    k : int, optional
    Returns
    -------
    ST : array
        If k = 0, ST segment amplitude values (milivolts).
    ST : int
        If k != 0, value obtained by using the ST segment expression for the given k value.

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    """
    if sm < 1 or sm > 150:
        raise Exception("Warning! sm is out of boundaries.")
    elif Kst < 0 or Kst > 110:
        raise Exception("Warning! Kst is out of boundaries.")
    else:
        if k == 0:
            k = np.arange(0, Kst, i)
            a = -S(i, As, Ks, Kcs, Ks - Kcs) * (k / sm) + S(i, As, Ks, Kcs, Ks - Kcs)
            ST = a.tolist()
        else:
            ST = -S(i, As, Ks, Kcs, Ks - Kcs) * (k / sm) + S(i, As, Ks, Kcs, Ks - Kcs)
    return ST


def T(i, As, Ks, Kcs, sm, Kst, At, Kt, k=0):
    """Generates the amplitude values of the T wave in the ECG signal.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameters introduced don't make sense in this context, an error will raise.
    Parameters
    ----------
    i  : int
        Sampling rate.
    As : int
        S wave amplitude (milivolts).
    Ks : int
        S wave width (miliseconds).
    Kcs : int
        Parameter which allows slight adjustment of S wave shape by cutting away a portion at the end.
    sm : int
        Slope parameter in the ST segment.
    Kst : int
        ST segment width (miliseconds).
    At : int
        1/2 of the T wave amplitude (milivolts).
    Kt : int
        T wave width (miliseconds).
    k : int, optional
    Returns
    -------
    T : array
        If k = 0, T wave amplitude values (milivolts).
    T : int
        If k != 0, value obtained by using the T wave expression for the given k value.

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    """
    if At < -0.5 or At > 1:
        raise Exception("Warning! At is out of boundaries.")
    elif Kt < 50 or Kt > 300:
        raise Exception("Warning! Kt is out of boundaries.")
    else:
        if k == 0:
            k = np.arange(0, Kt, i)
            a = (
                -At * np.cos((1.48 * np.pi * k + 15) / Kt)
                + At
                + St(i, As, Ks, Kcs, sm, Kst, Kst)
            )
            T = a.tolist()
        else:
            T = (
                -At * np.cos((1.48 * np.pi * k + 15) / Kt)
                + At
                + St(i, As, Ks, Kcs, sm, Kst, Kst)
            )
    return T


def I(i, As, Ks, Kcs, sm, Kst, At, Kt, si, Ki):
    """Generates the amplitude values of the final isoelectric segment (I segment) in the ECG signal.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameters introduced don't make sense in this context, an error will raise.
    Parameters
    ----------
    i  : int
        Sampling rate.
    As : int
        S wave amplitude (milivolts).
    Ks : int
        S wave width (miliseconds).
    Kcs : int
        Parameter which allows slight adjustment of S wave shape by cutting away a portion at the end.
    sm : int
        Slope parameter in the ST segment.
    Kst : int
        ST segment width (miliseconds).
    At : int
        1/2 of the T wave amplitude (milivolts).
    Kt : int
        T wave width (miliseconds).
    si : int
        Parameter for setting the transition slope between T wave and isoelectric line.
    Ki : int
        I segment width (miliseconds).
    Returns
    -------
    I_segment : array
        I segment amplitude values (milivolts).

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    """
    if si < 0 or si > 50:
        raise Exception("Warning! si is out of boundaries.")
    else:
        k = np.arange(0, Ki, i)
        a = T(i, As, Ks, Kcs, sm, Kst, At, Kt, Kt) * (si / (k + 10))
        I_segment = a.tolist()
    return I_segment


def ecg(
    Kb=130,
    Ap=0.2,
    Kp=100,
    Kpq=40,
    Aq=0.1,
    Kq1=25,
    Kq2=5,
    Ar=0.7,
    Kr=40,
    As=0.2,
    Ks=30,
    Kcs=5,
    sm=96,
    Kst=100,
    At=0.15,
    Kt=220,
    si=2,
    Ki=200,
    var=0.01,
    sampling_rate=10000,
):  # normal values by default
    """Concatenates the segments and waves to make an ECG signal. The default values are physiological.

    Follows the approach by Dolinský, Andráš, Michaeli and Grimaldi [Model03].

    If the parameters introduced aren't within physiological values (limits based on the website [ECGwaves]), a warning will raise.

    Parameters
    ----------
    Kb : int, optional
        B segment width (miliseconds).
    Ap : float, optional
        P wave amplitude (milivolts).
    Kp : int, optional
        P wave width (miliseconds).
    Kpq : int, optional
        PQ segment width (miliseconds).
    Aq : float, optional
        Q wave amplitude (milivolts).
    Kq1 : int, optional
        First 5/6 of the Q wave width (miliseconds).
    Kq2 : int, optional
        Last 1/6 of the Q wave width (miliseconds).
    Ar : float, optional
        R wave amplitude (milivolts).
    Kr : int, optional
        R wave width (miliseconds).
    As : float, optional
        S wave amplitude (milivolts).
    Ks : int, optional
        S wave width (miliseconds).
    Kcs : int, optional
        Parameter which allows slight adjustment of S wave shape by cutting away a portion at the end.
    sm : int, optional
        Slope parameter in the ST segment.
    Kst : int, optional
        ST segment width (miliseconds).
    At : float, optional
        1/2 of the T wave amplitude (milivolts).
    Kt : int, optional
        T wave width (miliseconds).
    si : int, optional
        Parameter for setting the transition slope between T wave and isoelectric line.
    Ki : int, optional
        I segment width (miliseconds).
    var : float, optional
        Value between 0.0 and 1.0 that adds variability to the obtained signal, by changing each parameter following a normal distribution with mean value `parameter_value` and std `var * parameter_value`.
    sampling_rate : int, optional
        Sampling frequency (Hz).

    Returns
    -------
    ecg : array
        Amplitude values of the ECG wave.
    t : array
        Time values accoring to the provided sampling rate.
    params : dict
        Input parameters of the function


    Example
    -------
    sampling_rate = 10000
    beats = 3
    noise_amplitude = 0.05

    ECGtotal = np.array([])
    for i in range(beats):
        ECGwave, _, _ = ecg(sampling_rate=sampling_rate, var=0.1)
        ECGtotal = np.concatenate((ECGtotal, ECGwave))
    t = np.arange(0, len(ECGtotal)) / sampling_rate

    # add powerline noise (50 Hz)
    noise = noise_amplitude * np.sin(50 * (2 * pi) * t)
    ECGtotal += noise

    plt.plot(t, ECGtotal)
    plt.xlabel("Time (ms)")
    plt.ylabel("Amplitude (mV)")
    plt.grid()
    plt.title("ECG")

    plt.show()

    References
    ----------
    .. [Model03] Pavol DOLINSKÝ, Imrich ANDRÁŠ, Linus MICHAELI, Domenico GRIMALDI,
       "MODEL FOR GENERATING SIMPLE SYNTHETIC ECG SIGNALS",
       Acta Electrotechnica et Informatica, Vol. 18, No. 3, 2018, 3–8
    .. [ECGwaves] https://ecgwaves.com/
    """
    if Kp > 120 and Ap >= 0.25:
        warnings.warn("P wave isn't within physiological values.")

    if Kq1 + Kq2 > 30 or Aq > 0.25 * Ar:
        warnings.warn("Q wave isn't within physiological values.")

    if 120 > Kp + Kpq or Kp + Kpq > 220:
        warnings.warn("PR interval isn't within physiological limits.")

    if Kq1 + Kq2 + Kr + Ks - Kcs > 120:
        warnings.warn("QRS complex duration isn't within physiological limits.")

    if Kq1 + Kq2 + Kr + Ks - Kcs + Kst + Kt > 450:
        warnings.warn("QT segment duration isn't within physiological limits for men.")

    if Kq1 + Kq2 + Kr + Ks - Kcs + Kst + Kt > 470:
        warnings.warn(
            "QT segment duration isn't within physiological limits for women."
        )

    if var < 0 or var > 1:
        raise TypeError("Variability value should be between 0.0 and 1.0")

    if var > 0:
        # change the parameter according to the provided variability
        nd = lambda x: np.random.normal(x, x * var)
        Kb = round(np.clip(nd(Kb), 0, 130))
        Ap = np.clip(nd(Ap), -0.2, 0.5)
        Kp = np.clip(nd(Kp), 10, 100)
        Kpq = round(np.clip(nd(Kpq), 0, 60))
        Aq = np.clip(nd(Aq), 0, 0.5)
        Kq1 = round(np.clip(nd(Kq1), 0, 70))
        Kq2 = round(np.clip(nd(Kq2), 0, 50))
        Ar = np.clip(nd(Ar), 0.5, 2)
        Kr = round(np.clip(nd(Kr), 10, 150))
        As = np.clip(nd(As), 0, 1)
        Ks = round(np.clip(nd(Ks), 10, 200))
        Kcs = round(np.clip(nd(Kcs), -5, 150))
        sm = round(np.clip(nd(sm), 1, 150))
        Kst = round(np.clip(nd(Kst), 0, 110))
        At = np.clip(nd(At), -0.5, 1)
        Kt = round(np.clip(nd(Kt), 50, 300))
        si = round(np.clip(nd(si), 0, 50))

    # variable i is the time between samples (in miliseconds)
    i = 1000 / sampling_rate
    l = int(1 / i)

    B_to_S = (
        B(l, Kb)
        + P(i, Ap, Kp)
        + Pq(l, Kpq)
        + Q1(i, Aq, Kq1)
        + Q2(i, Aq, Kq2)
        + R(i, Ar, Kr)
        + S(i, As, Ks, Kcs)
    )
    St_to_I = (
        St(i, As, Ks, Kcs, sm, Kst)
        + T(i, As, Ks, Kcs, sm, Kst, At, Kt)
        + I(i, As, Ks, Kcs, sm, Kst, At, Kt, si, Ki)
    )

    # The signal is filtered in two different sizes
    ECG1_filtered, n1 = st.smoother(B_to_S, size=50)
    ECG2_filtered, n2 = st.smoother(St_to_I, size=500)

    # The signal is concatenated
    ECGwave = np.concatenate((ECG1_filtered, ECG2_filtered))

    # Time array
    t = np.arange(0, len(ECGwave)) / sampling_rate

    # output
    params = {
        "Kb": 130,
        "Ap": 0.2,
        "Kp": 100,
        "Kpq": 40,
        "Aq": 0.1,
        "Kq1": 25,
        "Kq2": 5,
        "Ar": 0.7,
        "Kr": 40,
        "As": 0.2,
        "Ks": 30,
        "Kcs": 5,
        "sm": 96,
        "Kst": 100,
        "At": 0.15,
        "Kt": 220,
        "si": 2,
        "Ki": 200,
        "var": 0.01,
        "sampling_rate": 10000,
    }

    args = (ECGwave, t, params)
    names = ("ecg", "t", "params")

    return utils.ReturnTuple(args, names)



# -*- coding: utf-8 -*-
"""
biosppy
-------

A toolbox for biosignal processing written in Python.

:copyright: (c) 2015-2021 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# compat
from __future__ import absolute_import, division, print_function

# get version
from .__version__ import __version__

# allow lazy loading
from .signals import acc, abp, bvp, ppg, pcg, ecg, eda, eeg, emg, resp, tools
from .synthesizers import ecg
